import abc
from dataclasses import asdict, dataclass, field
from datetime import datetime
import logging
import re
import time
# Due to a failure in python 3.8
from typing import List

import dacite
from dataclasses_json import dataclass_json

from aprsd.utils import counter


LOG = logging.getLogger("APRSD")

PACKET_TYPE_MESSAGE = "message"
PACKET_TYPE_ACK = "ack"
PACKET_TYPE_REJECT = "reject"
PACKET_TYPE_MICE = "mic-e"
PACKET_TYPE_WX = "weather"
PACKET_TYPE_OBJECT = "object"
PACKET_TYPE_UNKNOWN = "unknown"
PACKET_TYPE_STATUS = "status"
PACKET_TYPE_BEACON = "beacon"
PACKET_TYPE_THIRDPARTY = "thirdparty"
PACKET_TYPE_UNCOMPRESSED = "uncompressed"

NO_DATE = datetime(1900, 10, 24)


def _init_timestamp():
    """Build a unix style timestamp integer"""
    return int(round(time.time()))


def _init_send_time():
    # We have to use a datetime here, or the json encoder
    # Fails on a NoneType.
    return NO_DATE


def _init_msgNo():    # noqa: N802
    """For some reason __post__init doesn't get called.

    So in order to initialize the msgNo field in the packet
    we use this workaround.
    """
    c = counter.PacketCounter()
    c.increment()
    return c.value


def factory_from_dict(packet_dict):
    pkt_type = get_packet_type(packet_dict)
    if pkt_type:
        cls = TYPE_LOOKUP[pkt_type]
        return cls.from_dict(packet_dict)


def factory_from_json(packet_dict):
    pkt_type = get_packet_type(packet_dict)
    if pkt_type:
        return TYPE_LOOKUP[pkt_type].from_json(packet_dict)


@dataclass_json
@dataclass(unsafe_hash=True)
class Packet(metaclass=abc.ABCMeta):
    from_call: str = field(default=None)
    to_call: str = field(default=None)
    addresse: str = field(default=None)
    format: str = field(default=None)
    msgNo: str = field(default_factory=_init_msgNo)   # noqa: N815
    packet_type: str = field(default=None)
    timestamp: float = field(default_factory=_init_timestamp, compare=False, hash=False)
    # Holds the raw text string to be sent over the wire
    # or holds the raw string from input packet
    raw: str = field(default=None, compare=False, hash=False)
    raw_dict: dict = field(repr=False, default_factory=lambda: {}, compare=False, hash=False)
    # Built by calling prepare().  raw needs this built first.
    payload: str = field(default=None)

    # Fields related to sending packets out
    send_count: int = field(repr=False, default=0, compare=False, hash=False)
    retry_count: int = field(repr=False, default=3, compare=False, hash=False)
    # last_send_time: datetime = field(
    #    metadata=dc_json_config(
    #        encoder=datetime.isoformat,
    #        decoder=datetime.fromisoformat,
    #    ),
    #    repr=True,
    #    default_factory=_init_send_time,
    #    compare=False,
    #    hash=False
    # )
    last_send_time: float = field(repr=False, default=0, compare=False, hash=False)
    last_send_attempt: int = field(repr=False, default=0, compare=False, hash=False)

    # Do we allow this packet to be saved to send later?
    allow_delay: bool = field(repr=False, default=True, compare=False, hash=False)
    path: List[str] = field(default_factory=list, compare=False, hash=False)
    via: str = field(default=None, compare=False, hash=False)

    def __post__init__(self):
        LOG.warning(f"POST INIT {self}")

    @property
    def json(self):
        """
        get the json formated string
        """
        return self.to_json()

    def get(self, key, default=None):
        """Emulate a getter on a dict."""
        if hasattr(self, key):
            return getattr(self, key)
        else:
            return default

    @property
    def key(self):
        """Build a key for finding this packet in a dict."""
        return f"{self.from_call}:{self.addresse}:{self.msgNo}"

    def update_timestamp(self):
        self.timestamp = _init_timestamp()

    def prepare(self):
        """Do stuff here that is needed prior to sending over the air."""
        # now build the raw message for sending
        self._build_payload()
        self._build_raw()

    def _build_payload(self):
        """The payload is the non headers portion of the packet."""
        msg = self._filter_for_send().rstrip("\n")
        self.payload = (
            f":{self.to_call.ljust(9)}"
            f":{msg}"
        )

    def _build_raw(self):
        """Build the self.raw which is what is sent over the air."""
        self.raw = "{}>APZ100:{}".format(
            self.from_call,
            self.payload,
        )

    @staticmethod
    def factory(raw_packet):
        """Factory method to create a packet from a raw packet string."""
        raw = raw_packet
        raw["raw_dict"] = raw.copy()
        translate_fields = {
            "from": "from_call",
            "to": "to_call",
        }
        # First translate some fields
        for key in translate_fields:
            if key in raw:
                raw[translate_fields[key]] = raw[key]
                del raw[key]

        if "addresse" in raw:
            raw["to_call"] = raw["addresse"]

        packet_type = get_packet_type(raw)
        raw["packet_type"] = packet_type
        class_name = TYPE_LOOKUP[packet_type]
        if packet_type == PACKET_TYPE_THIRDPARTY:
            # We have an encapsulated packet!
            # So we need to decode it and return the inner packet
            # as the packet we are going to process.
            # This is a recursive call to the factory
            subpacket_raw = raw["subpacket"]
            subpacket = Packet.factory(subpacket_raw)
            del raw["subpacket"]
            # raw["subpacket"] = subpacket
            packet = dacite.from_dict(data_class=class_name, data=raw)
            packet.subpacket = subpacket
            return packet

        if packet_type == PACKET_TYPE_UNKNOWN:
            # Try and figure it out here
            if "latitude" in raw:
                class_name = GPSPacket

        if packet_type == PACKET_TYPE_WX:
            # the weather information is in a dict
            # this brings those values out to the outer dict

            for key in raw["weather"]:
                raw[key] = raw["weather"][key]

            # If we have the broken aprslib, then we need to
            # Convert the course and speed to wind_speed and wind_direction
            # aprslib issue #80
            # https://github.com/rossengeorgiev/aprs-python/issues/80
            # Wind speed and course is option in the SPEC.
            # For some reason aprslib multiplies the speed by 1.852.
            if "wind_speed" not in raw and "wind_direction" not in raw:
                # Most likely this is the broken aprslib
                # So we need to convert the wind_gust speed
                raw["wind_gust"] = round(raw.get("wind_gust", 0) / 0.44704, 3)
            if "wind_speed" not in raw:
                wind_speed = raw.get("speed")
                if wind_speed:
                    raw["wind_speed"] = round(wind_speed / 1.852, 3)
                    raw["weather"]["wind_speed"] = raw["wind_speed"]
                if "speed" in raw:
                    del raw["speed"]
                # Let's adjust the rain numbers as well, since it's wrong
                raw["rain_1h"] = round((raw.get("rain_1h", 0) / .254) * .01, 3)
                raw["weather"]["rain_1h"] = raw["rain_1h"]
                raw["rain_24h"] = round((raw.get("rain_24h", 0) / .254) * .01, 3)
                raw["weather"]["rain_24h"] = raw["rain_24h"]
                raw["rain_since_midnight"] = round((raw.get("rain_since_midnight", 0) / .254) * .01, 3)
                raw["weather"]["rain_since_midnight"] = raw["rain_since_midnight"]

            if "wind_direction" not in raw:
                wind_direction = raw.get("course")
                if wind_direction:
                    raw["wind_direction"] = wind_direction
                    raw["weather"]["wind_direction"] = raw["wind_direction"]
                if "course" in raw:
                    del raw["course"]

        return dacite.from_dict(data_class=class_name, data=raw)

    def log(self, header=None):
        """LOG a packet to the logfile."""
        asdict(self)
        log_list = ["\n"]
        name = self.__class__.__name__
        if header:
            if "tx" in header.lower():
                log_list.append(
                    f"{header}________({name}  "
                    f"TX:{self.send_count+1} of {self.retry_count})",
                )
            else:
                log_list.append(f"{header}________({name})")
        # log_list.append(f"  Packet  : {self.__class__.__name__}")
        log_list.append(f"  Raw     : {self.raw}")
        if self.to_call:
            log_list.append(f"  To      : {self.to_call}")
        if self.from_call:
            log_list.append(f"  From    : {self.from_call}")
        if hasattr(self, "path") and self.path:
            log_list.append(f"  Path    : {'=>'.join(self.path)}")
        if hasattr(self, "via") and self.via:
            log_list.append(f"  VIA     : {self.via}")

        elif isinstance(self, MessagePacket):
            log_list.append(f"  Message : {self.message_text}")

        if hasattr(self, "comment") and self.comment:
            log_list.append(f"  Comment : {self.comment}")

        if self.msgNo:
            log_list.append(f"  Msg #   : {self.msgNo}")
        log_list.append(f"{header}________({name})")

        LOG.info("\n".join(log_list))
        LOG.debug(repr(self))

    def _filter_for_send(self) -> str:
        """Filter and format message string for FCC."""
        # max?  ftm400 displays 64, raw msg shows 74
        # and ftm400-send is max 64.  setting this to
        # 67 displays 64 on the ftm400. (+3 {01 suffix)
        # feature req: break long ones into two msgs
        message = self.raw[:67]
        # We all miss George Carlin
        return re.sub("fuck|shit|cunt|piss|cock|bitch", "****", message)

    def __str__(self):
        """Show the raw version of the packet"""
        self.prepare()
        return self.raw

    def __repr__(self):
        """Build the repr version of the packet."""
        repr = (
            f"{self.__class__.__name__}:"
            f" From: {self.from_call}  "
            " To: "
        )

        return repr


@dataclass(unsafe_hash=True)
class AckPacket(Packet):
    response: str = field(default=None)

    def __post__init__(self):
        if self.response:
            LOG.warning("Response set!")

    def _build_payload(self):
        self.payload = f":{self.to_call.ljust(9)}:ack{self.msgNo}"


@dataclass(unsafe_hash=True)
class RejectPacket(Packet):
    response: str = field(default=None)

    def __post__init__(self):
        if self.response:
            LOG.warning("Response set!")

    def _build_payload(self):
        self.payload = f":{self.to_call.ljust(9)} :rej{self.msgNo}"


@dataclass_json
@dataclass(unsafe_hash=True)
class MessagePacket(Packet):
    message_text: str = field(default=None)

    def _filter_for_send(self) -> str:
        """Filter and format message string for FCC."""
        # max?  ftm400 displays 64, raw msg shows 74
        # and ftm400-send is max 64.  setting this to
        # 67 displays 64 on the ftm400. (+3 {01 suffix)
        # feature req: break long ones into two msgs
        message = self.message_text[:67]
        # We all miss George Carlin
        return re.sub("fuck|shit|cunt|piss|cock|bitch", "****", message)

    def _build_payload(self):
        self.payload = ":{}:{}{{{}".format(
            self.to_call.ljust(9),
            self._filter_for_send().rstrip("\n"),
            str(self.msgNo),
        )


@dataclass(unsafe_hash=True)
class StatusPacket(Packet):
    status: str = field(default=None)
    messagecapable: bool = field(default=False)
    comment: str = field(default=None)

    def _build_payload(self):
        raise NotImplementedError


@dataclass(unsafe_hash=True)
class GPSPacket(Packet):
    latitude: float = field(default=0.00)
    longitude: float = field(default=0.00)
    altitude: float = field(default=0.00)
    rng: float = field(default=0.00)
    posambiguity: int = field(default=0)
    comment: str = field(default=None)
    symbol: str = field(default="l")
    symbol_table: str = field(default="/")

    def decdeg2dms(self, degrees_decimal):
        is_positive = degrees_decimal >= 0
        degrees_decimal = abs(degrees_decimal)
        minutes, seconds = divmod(degrees_decimal * 3600, 60)
        degrees, minutes = divmod(minutes, 60)
        degrees = degrees if is_positive else -degrees

        degrees = str(int(degrees)).replace("-", "0")
        minutes = str(int(minutes)).replace("-", "0")
        seconds = str(int(round(seconds * 0.01, 2) * 100))

        return {"degrees": degrees, "minutes": minutes, "seconds": seconds}

    def decdeg2dmm_m(self, degrees_decimal):
        is_positive = degrees_decimal >= 0
        degrees_decimal = abs(degrees_decimal)
        minutes, seconds = divmod(degrees_decimal * 3600, 60)
        degrees, minutes = divmod(minutes, 60)
        degrees = degrees if is_positive else -degrees

        degrees = abs(int(degrees))
        minutes = int(round(minutes + (seconds / 60), 2))
        hundredths = round(seconds / 60, 2)

        return {
            "degrees": degrees, "minutes": minutes, "seconds": seconds,
            "hundredths": hundredths,
        }

    def convert_latitude(self, degrees_decimal):
        det = self.decdeg2dmm_m(degrees_decimal)
        if degrees_decimal > 0:
            direction = "N"
        else:
            direction = "S"

        degrees = str(det.get("degrees")).zfill(2)
        minutes = str(det.get("minutes")).zfill(2)
        seconds = det.get("seconds")
        hun = det.get("hundredths")
        hundredths = f"{hun:.2f}".split(".")[1]

        LOG.debug(
            f"LAT degress {degrees}  minutes {str(minutes)} "
            f"seconds {seconds} hundredths {hundredths} direction {direction}",
        )

        lat = f"{degrees}{str(minutes)}.{hundredths}{direction}"
        return lat

    def convert_longitude(self, degrees_decimal):
        det = self.decdeg2dmm_m(degrees_decimal)
        if degrees_decimal > 0:
            direction = "E"
        else:
            direction = "W"

        degrees = str(det.get("degrees")).zfill(3)
        minutes = str(det.get("minutes")).zfill(2)
        seconds = det.get("seconds")
        hun = det.get("hundredths")
        hundredths = f"{hun:.2f}".split(".")[1]

        LOG.debug(
            f"LON degress {degrees}  minutes {str(minutes)} "
            f"seconds {seconds} hundredths {hundredths} direction {direction}",
        )

        lon = f"{degrees}{str(minutes)}.{hundredths}{direction}"
        return lon

    def _build_time_zulu(self):
        """Build the timestamp in UTC/zulu."""
        if self.timestamp:
            local_dt = datetime.fromtimestamp(self.timestamp)
        else:
            local_dt = datetime.now()
            self.timestamp = datetime.timestamp(local_dt)

        utc_offset_timedelta = datetime.utcnow() - local_dt
        result_utc_datetime = local_dt + utc_offset_timedelta
        time_zulu = result_utc_datetime.strftime("%d%H%M")
        return time_zulu

    def _build_payload(self):
        """The payload is the non headers portion of the packet."""
        time_zulu = self._build_time_zulu()
        lat = self.latitude
        long = self.longitude
        self.payload = (
            f"@{time_zulu}z{lat}{self.symbol_table}"
            f"{long}{self.symbol}"
        )

        if self.comment:
            self.payload = f"{self.payload}{self.comment}"

    def _build_raw(self):
        self.raw = (
            f"{self.from_call}>{self.to_call},WIDE2-1:"
            f"{self.payload}"
        )


@dataclass(unsafe_hash=True)
class BeaconPacket(GPSPacket):
    def _build_payload(self):
        """The payload is the non headers portion of the packet."""
        time_zulu = self._build_time_zulu()
        lat = self.convert_latitude(self.latitude)
        long = self.convert_longitude(self.longitude)

        self.payload = (
            f"@{time_zulu}z{lat}{self.symbol_table}"
            f"{long}{self.symbol}APRSD Beacon"
        )

    def _build_raw(self):
        self.raw = (
            f"{self.from_call}>APZ100:"
            f"{self.payload}"
        )


@dataclass
class MicEPacket(GPSPacket):
    messagecapable: bool = False
    mbits: str = None
    mtype: str = None
    # in MPH
    speed: float = 0.00
    # 0 to 360
    course: int = 0

    def _build_payload(self):
        raise NotImplementedError


@dataclass
class ObjectPacket(GPSPacket):
    alive: bool = True
    raw_timestamp: str = None
    symbol: str = field(default="r")
    # in MPH
    speed: float = 0.00
    # 0 to 360
    course: int = 0

    def _build_payload(self):
        time_zulu = self._build_time_zulu()
        lat = self.convert_latitude(self.latitude)
        long = self.convert_longitude(self.longitude)

        self.payload = (
            f"*{time_zulu}z{lat}{self.symbol_table}"
            f"{long}{self.symbol}"
        )

        if self.comment:
            self.payload = f"{self.payload}{self.comment}"

    def _build_raw(self):
        """
        REPEAT builds packets like
        reply = "{}>APZ100:;{:9s}*{}z{}r{:.3f}MHz {} {}".format(
                fromcall, callsign, time_zulu, latlon, freq, uplink_tone, offset,
            )
        where fromcall is the callsign that is sending the packet
        callsign is the station callsign for the object
        The frequency, uplink_tone, offset is part of the comment
        """

        self.raw = (
            f"{self.from_call}>APZ100:;{self.to_call:9s}"
            f"{self.payload}"
        )


@dataclass()
class WeatherPacket(GPSPacket):
    symbol: str = "_"
    wind_speed: float = 0.00
    wind_direction: int = 0
    wind_gust: float = 0.00
    temperature: float = 0.00
    # in inches.  1.04 means 1.04 inches
    rain_1h: float = 0.00
    rain_24h: float = 0.00
    rain_since_midnight: float = 0.00
    humidity: int = 0
    pressure: float = 0.00
    comment: str = None

    def _build_payload(self):
        """Build an uncompressed weather packet

        Format =

       _CSE/SPDgXXXtXXXrXXXpXXXPXXXhXXbXXXXX%type NEW FORMAT APRS793 June 97
                                                  NOT BACKWARD COMPATIBLE


        Where: CSE/SPD is wind direction and sustained 1 minute speed
        t is in degrees F

        r is Rain per last 60 minutes
            1.04 inches of rain will show as r104
        p is precipitation per last 24 hours (sliding 24 hour window)
        P is precip per last 24 hours since midnight
        b is Baro in tenths of a mb
        h is humidity in percent. 00=100
        g is Gust (peak winds in last 5 minutes)
        # is the raw rain counter for remote WX stations
        See notes on remotes below
        % shows software type d=Dos, m=Mac, w=Win, etc
        type shows type of WX instrument

        """
        time_zulu = self._build_time_zulu()

        contents = [
            f"@{time_zulu}z{self.latitude}{self.symbol_table}",
            f"{self.longitude}{self.symbol}",
            f"{self.wind_direction:03d}",
            # Speed = sustained 1 minute wind speed in mph
            f"{self.symbol_table}", f"{self.wind_speed:03.0f}",
            # wind gust (peak wind speed in mph in the last 5 minutes)
            f"g{self.wind_gust:03.0f}",
            # Temperature in degrees F
            f"t{self.temperature:03.0f}",
            # Rainfall (in hundredths of an inch) in the last hour
            f"r{self.rain_1h*100:03.0f}",
            # Rainfall (in hundredths of an inch) in last 24 hours
            f"p{self.rain_24h*100:03.0f}",
            # Rainfall (in hundredths of an inch) since midnigt
            f"P{self.rain_since_midnight*100:03.0f}",
            # Humidity
            f"h{self.humidity:02d}",
            # Barometric pressure (in tenths of millibars/tenths of hPascal)
            f"b{self.pressure:05.0f}",
        ]
        if self.comment:
            contents.append(self.comment)
        self.payload = "".join(contents)

    def _build_raw(self):

        self.raw = (
            f"{self.from_call}>{self.to_call},WIDE1-1,WIDE2-1:"
            f"{self.payload}"
        )


class ThirdParty(Packet):
    # Holds the encapsulated packet
    subpacket: Packet = field(default=None, compare=True, hash=False)

    def __repr__(self):
        """Build the repr version of the packet."""
        repr_str = (
            f"{self.__class__.__name__}:"
            f" From: {self.from_call}  "
            f" To: {self.to_call}  "
            f" Subpacket: {repr(self.subpacket)}"
        )

        return repr_str


TYPE_LOOKUP = {
    PACKET_TYPE_WX: WeatherPacket,
    PACKET_TYPE_MESSAGE: MessagePacket,
    PACKET_TYPE_ACK: AckPacket,
    PACKET_TYPE_REJECT: RejectPacket,
    PACKET_TYPE_MICE: MicEPacket,
    PACKET_TYPE_OBJECT: ObjectPacket,
    PACKET_TYPE_STATUS: StatusPacket,
    PACKET_TYPE_BEACON: GPSPacket,
    PACKET_TYPE_UNKNOWN: Packet,
    PACKET_TYPE_THIRDPARTY: ThirdParty,
}


def get_packet_type(packet: dict):
    """Decode the packet type from the packet."""

    pkt_format = packet.get("format", None)
    msg_response = packet.get("response", None)
    packet_type = PACKET_TYPE_UNKNOWN
    if pkt_format == "message" and msg_response == "ack":
        packet_type = PACKET_TYPE_ACK
    elif pkt_format == "message" and msg_response == "rej":
        packet_type = PACKET_TYPE_REJECT
    elif pkt_format == "message":
        packet_type = PACKET_TYPE_MESSAGE
    elif pkt_format == "mic-e":
        packet_type = PACKET_TYPE_MICE
    elif pkt_format == "object":
        packet_type = PACKET_TYPE_OBJECT
    elif pkt_format == "status":
        packet_type = PACKET_TYPE_STATUS
    elif pkt_format == PACKET_TYPE_BEACON:
        packet_type = PACKET_TYPE_BEACON
    elif pkt_format == PACKET_TYPE_UNCOMPRESSED:
        if packet.get("symbol", None) == "_":
            packet_type = PACKET_TYPE_WX
    elif pkt_format == PACKET_TYPE_THIRDPARTY:
        packet_type = PACKET_TYPE_THIRDPARTY

    if packet_type == PACKET_TYPE_UNKNOWN:
        if "latitude" in packet:
            packet_type = PACKET_TYPE_BEACON
    return packet_type


def is_message_packet(packet):
    return get_packet_type(packet) == PACKET_TYPE_MESSAGE


def is_ack_packet(packet):
    return get_packet_type(packet) == PACKET_TYPE_ACK


def is_mice_packet(packet):
    return get_packet_type(packet) == PACKET_TYPE_MICE



import datetime
import logging
import threading

from oslo_config import cfg
import wrapt

from aprsd.utils import objectstore


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class SeenList(objectstore.ObjectStoreMixin):
    """Global callsign seen list."""

    _instance = None
    lock = threading.Lock()
    data: dict = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._init_store()
            cls._instance.data = {}
        return cls._instance

    @wrapt.synchronized(lock)
    def update_seen(self, packet):
        callsign = None
        if packet.from_call:
            callsign = packet.from_call
        else:
            LOG.warning(f"Can't find FROM in packet {packet}")
            return
        if callsign not in self.data:
            self.data[callsign] = {
                "last": None,
                "count": 0,
            }
        self.data[callsign]["last"] = str(datetime.datetime.now())
        self.data[callsign]["count"] += 1



import datetime
import importlib.metadata as imp
import io
import json
import logging
import time

import flask
from flask import Flask
from flask_httpauth import HTTPBasicAuth
from oslo_config import cfg, generator
import socketio
from werkzeug.security import check_password_hash

import aprsd
from aprsd import cli_helper, client, conf, packets, plugin, threads
from aprsd.log import log
from aprsd.rpc import client as aprsd_rpc_client


CONF = cfg.CONF
LOG = logging.getLogger("gunicorn.access")

auth = HTTPBasicAuth()
users = {}
app = Flask(
    "aprsd",
    static_url_path="/static",
    static_folder="web/admin/static",
    template_folder="web/admin/templates",
)
bg_thread = None
app.config["SECRET_KEY"] = "secret!"


# HTTPBasicAuth doesn't work on a class method.
# This has to be out here.  Rely on the APRSDFlask
# class to initialize the users from the config
@auth.verify_password
def verify_password(username, password):
    global users

    if username in users and check_password_hash(users.get(username), password):
        return username


def _stats():
    track = aprsd_rpc_client.RPCClient().get_packet_track()
    now = datetime.datetime.now()

    time_format = "%m-%d-%Y %H:%M:%S"

    stats_dict = aprsd_rpc_client.RPCClient().get_stats_dict()
    if not stats_dict:
        stats_dict = {
            "aprsd": {},
            "aprs-is": {"server": ""},
            "messages": {
                "sent": 0,
                "received": 0,
            },
            "email": {
                "sent": 0,
                "received": 0,
            },
            "seen_list": {
                "sent": 0,
                "received": 0,
            },
        }

    # Convert the watch_list entries to age
    wl = aprsd_rpc_client.RPCClient().get_watch_list()
    new_list = {}
    if wl:
        for call in wl.get_all():
            # call_date = datetime.datetime.strptime(
            #    str(wl.last_seen(call)),
            #    "%Y-%m-%d %H:%M:%S.%f",
            # )

            # We have to convert the RingBuffer to a real list
            # so that json.dumps works.
            # pkts = []
            # for pkt in wl.get(call)["packets"].get():
            #     pkts.append(pkt)

            new_list[call] = {
                "last": wl.age(call),
                # "packets": pkts
            }

    stats_dict["aprsd"]["watch_list"] = new_list
    packet_list = aprsd_rpc_client.RPCClient().get_packet_list()
    rx = tx = 0
    types = {}
    if packet_list:
        rx = packet_list.total_rx()
        tx = packet_list.total_tx()
        types_copy = packet_list.types.copy()

        for key in types_copy:
            types[str(key)] = dict(types_copy[key])

    stats_dict["packets"] = {
        "sent": tx,
        "received": rx,
        "types": types,
    }
    if track:
        size_tracker = len(track)
    else:
        size_tracker = 0

    result = {
        "time": now.strftime(time_format),
        "size_tracker": size_tracker,
        "stats": stats_dict,
    }

    return result


@app.route("/stats")
def stats():
    LOG.debug("/stats called")
    return json.dumps(_stats())


@app.route("/")
def index():
    stats = _stats()
    wl = aprsd_rpc_client.RPCClient().get_watch_list()
    if wl and wl.is_enabled():
        watch_count = len(wl)
        watch_age = wl.max_delta()
    else:
        watch_count = 0
        watch_age = 0

    sl = aprsd_rpc_client.RPCClient().get_seen_list()
    if sl:
        seen_count = len(sl)
    else:
        seen_count = 0

    pm = plugin.PluginManager()
    plugins = pm.get_plugins()
    plugin_count = len(plugins)

    if CONF.aprs_network.enabled:
        transport = "aprs-is"
        aprs_connection = (
            "APRS-IS Server: <a href='http://status.aprs2.net' >"
            "{}</a>".format(stats["stats"]["aprs-is"]["server"])
        )
    else:
        # We might be connected to a KISS socket?
        if client.KISSClient.kiss_enabled():
            transport = client.KISSClient.transport()
            if transport == client.TRANSPORT_TCPKISS:
                aprs_connection = (
                    "TCPKISS://{}:{}".format(
                        CONF.kiss_tcp.host,
                        CONF.kiss_tcp.port,
                    )
                )
            elif transport == client.TRANSPORT_SERIALKISS:
                aprs_connection = (
                    "SerialKISS://{}@{} baud".format(
                        CONF.kiss_serial.device,
                        CONF.kiss_serial.baudrate,
                    )
                )

    stats["transport"] = transport
    stats["aprs_connection"] = aprs_connection
    entries = conf.conf_to_dict()

    return flask.render_template(
        "index.html",
        initial_stats=stats,
        aprs_connection=aprs_connection,
        callsign=CONF.callsign,
        version=aprsd.__version__,
        config_json=json.dumps(
            entries, indent=4,
            sort_keys=True, default=str,
        ),
        watch_count=watch_count,
        watch_age=watch_age,
        seen_count=seen_count,
        plugin_count=plugin_count,
        # oslo_out=generate_oslo()
    )


@auth.login_required
def messages():
    track = packets.PacketTrack()
    msgs = []
    for id in track:
        LOG.info(track[id].dict())
        msgs.append(track[id].dict())

    return flask.render_template("messages.html", messages=json.dumps(msgs))


@auth.login_required
@app.route("/packets")
def get_packets():
    LOG.debug("/packets called")
    packet_list = aprsd_rpc_client.RPCClient().get_packet_list()
    if packet_list:
        tmp_list = []
        pkts = packet_list.copy()
        for key in pkts:
            pkt = packet_list.get(key)
            if pkt:
                tmp_list.append(pkt.json)

        return json.dumps(tmp_list)
    else:
        return json.dumps([])


@auth.login_required
@app.route("/plugins")
def plugins():
    LOG.debug("/plugins called")
    pm = plugin.PluginManager()
    pm.reload_plugins()

    return "reloaded"


def _get_namespaces():
    args = []

    all = imp.entry_points()
    selected = []
    if "oslo.config.opts" in all:
        for x in all["oslo.config.opts"]:
            if x.group == "oslo.config.opts":
                selected.append(x)
    for entry in selected:
        if "aprsd" in entry.name:
            args.append("--namespace")
            args.append(entry.name)

    return args


def generate_oslo():
    CONF.namespace = _get_namespaces()
    string_out = io.StringIO()
    generator.generate(CONF, string_out)
    return string_out.getvalue()


@auth.login_required
@app.route("/oslo")
def oslo():
    return generate_oslo()


@auth.login_required
@app.route("/save")
def save():
    """Save the existing queue to disk."""
    track = packets.PacketTrack()
    track.save()
    return json.dumps({"messages": "saved"})


class LogUpdateThread(threads.APRSDThread):

    def __init__(self):
        super().__init__("LogUpdate")

    def loop(self):
        if sio:
            log_entries = aprsd_rpc_client.RPCClient().get_log_entries()

            if log_entries:
                LOG.info(f"Sending log entries! {len(log_entries)}")
                for entry in log_entries:
                    sio.emit(
                        "log_entry", entry,
                        namespace="/logs",
                    )
        time.sleep(5)
        return True


class LoggingNamespace(socketio.Namespace):
    log_thread = None

    def on_connect(self, sid, environ):
        global sio
        LOG.debug(f"LOG on_connect {sid}")
        sio.emit(
            "connected", {"data": "/logs Connected"},
            namespace="/logs",
        )
        self.log_thread = LogUpdateThread()
        self.log_thread.start()

    def on_disconnect(self, sid):
        LOG.debug(f"LOG Disconnected {sid}")
        if self.log_thread:
            self.log_thread.stop()


def init_app(config_file=None, log_level=None):
    default_config_file = cli_helper.DEFAULT_CONFIG_FILE
    if not config_file:
        config_file = default_config_file

    CONF(
        [], project="aprsd", version=aprsd.__version__,
        default_config_files=[config_file],
    )

    if not log_level:
        log_level = CONF.logging.log_level

    return log_level


if __name__ == "__main__":
    async_mode = "threading"
    sio = socketio.Server(logger=True, async_mode=async_mode)
    app.wsgi_app = socketio.WSGIApp(sio, app.wsgi_app)
    log_level = init_app(log_level="DEBUG")
    log.setup_logging(app, log_level)
    sio.register_namespace(LoggingNamespace("/logs"))
    CONF.log_opt_values(LOG, logging.DEBUG)
    app.run(
        threaded=True,
        debug=False,
        port=CONF.admin.web_port,
        host=CONF.admin.web_ip,
    )


if __name__ == "uwsgi_file_aprsd_wsgi":
    # Start with
    # uwsgi --http :8000 --gevent 1000 --http-websockets --master -w aprsd.wsgi --callable app

    async_mode = "gevent_uwsgi"
    sio = socketio.Server(logger=True, async_mode=async_mode)
    app.wsgi_app = socketio.WSGIApp(sio, app.wsgi_app)
    log_level = init_app(
        log_level="DEBUG",
        config_file="/config/aprsd.conf",
        # Commented out for local development.
        # config_file=cli_helper.DEFAULT_CONFIG_FILE
    )
    log.setup_logging(app, log_level)
    sio.register_namespace(LoggingNamespace("/logs"))
    CONF.log_opt_values(LOG, logging.DEBUG)


if __name__ == "aprsd.wsgi":
    # set async_mode to 'threading', 'eventlet', 'gevent' or 'gevent_uwsgi' to
    # force a mode else, the best mode is selected automatically from what's
    # installed
    async_mode = "gevent_uwsgi"
    sio = socketio.Server(logger=True, async_mode=async_mode)
    app.wsgi_app = socketio.WSGIApp(sio, app.wsgi_app)

    log_level = init_app(
        log_level="DEBUG",
        config_file="/config/aprsd.conf",
        # config_file=cli_helper.DEFAULT_CONFIG_FILE,
    )
    log.setup_logging(app, log_level)
    sio.register_namespace(LoggingNamespace("/logs"))
    CONF.log_opt_values(LOG, logging.DEBUG)



from multiprocessing import RawValue
import threading

import wrapt


class PacketCounter:
    """
    Global Packet id counter class.

    This is a singleton based class that keeps
    an incrementing counter for all packets to
    be sent.  All new Packet objects gets a new
    message id, which is the next number available
    from the PacketCounter.

    """

    _instance = None
    max_count = 9999
    lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        """Make this a singleton class."""
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance.val = RawValue("i", 1)
        return cls._instance

    @wrapt.synchronized(lock)
    def increment(self):
        if self.val.value == self.max_count:
            self.val.value = 1
        else:
            self.val.value += 1

    @property
    @wrapt.synchronized(lock)
    def value(self):
        return str(self.val.value)

    @wrapt.synchronized(lock)
    def __repr__(self):
        return str(self.val.value)

    @wrapt.synchronized(lock)
    def __str__(self):
        return str(self.val.value)



import logging

from ax253 import Frame
import kiss
from oslo_config import cfg

from aprsd import conf  # noqa
from aprsd.packets import core
from aprsd.utils import trace


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class KISS3Client:
    path = []

    def __init__(self):
        self.setup()

    def is_alive(self):
        return True

    def setup(self):
        # we can be TCP kiss or Serial kiss
        if CONF.kiss_serial.enabled:
            LOG.debug(
                "KISS({}) Serial connection to {}".format(
                    kiss.__version__,
                    CONF.kiss_serial.device,
                ),
            )
            self.kiss = kiss.SerialKISS(
                port=CONF.kiss_serial.device,
                speed=CONF.kiss_serial.baudrate,
                strip_df_start=True,
            )
            self.path = CONF.kiss_serial.path
        elif CONF.kiss_tcp.enabled:
            LOG.debug(
                "KISS({}) TCP Connection to {}:{}".format(
                    kiss.__version__,
                    CONF.kiss_tcp.host,
                    CONF.kiss_tcp.port,
                ),
            )
            self.kiss = kiss.TCPKISS(
                host=CONF.kiss_tcp.host,
                port=CONF.kiss_tcp.port,
                strip_df_start=True,
            )
            self.path = CONF.kiss_tcp.path

        LOG.debug("Starting KISS interface connection")
        self.kiss.start()

    @trace.trace
    def stop(self):
        try:
            self.kiss.stop()
            self.kiss.loop.call_soon_threadsafe(
                self.kiss.protocol.transport.close,
            )
        except Exception as ex:
            LOG.exception(ex)

    def set_filter(self, filter):
        # This does nothing right now.
        pass

    def parse_frame(self, frame_bytes):
        try:
            frame = Frame.from_bytes(frame_bytes)
            # Now parse it with aprslib
            kwargs = {
                "frame": frame,
            }
            self._parse_callback(**kwargs)
        except Exception as ex:
            LOG.error("Failed to parse bytes received from KISS interface.")
            LOG.exception(ex)

    def consumer(self, callback, blocking=False, immortal=False, raw=False):
        LOG.debug("Start blocking KISS consumer")
        self._parse_callback = callback
        self.kiss.read(callback=self.parse_frame, min_frames=None)
        LOG.debug(f"END blocking KISS consumer {self.kiss}")

    def send(self, packet):
        """Send an APRS Message object."""

        payload = None
        path = self.path
        if isinstance(packet, core.Packet):
            packet.prepare()
            payload = packet.payload.encode("US-ASCII")
            if packet.path:
                path = packet.path
        else:
            msg_payload = f"{packet.raw}{{{str(packet.msgNo)}"
            payload = (
                ":{:<9}:{}".format(
                    packet.to_call,
                    msg_payload,
                )
            ).encode("US-ASCII")

        LOG.debug(
            f"KISS Send '{payload}' TO '{packet.to_call}' From "
            f"'{packet.from_call}' with PATH '{path}'",
        )
        frame = Frame.ui(
            destination="APZ100",
            source=packet.from_call,
            path=path,
            info=payload,
        )
        self.kiss.write(frame)



import datetime
import logging
import re

from oslo_config import cfg

from aprsd import packets, plugin
from aprsd.packets import tracker
from aprsd.utils import trace


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class QueryPlugin(plugin.APRSDRegexCommandPluginBase):
    """Query command."""

    command_regex = r"^\!.*"
    command_name = "query"
    short_description = "APRSD Owner command to query messages in the MsgTrack"

    def setup(self):
        """Do any plugin setup here."""
        if not CONF.query_plugin.callsign:
            LOG.error("Config query_plugin.callsign not set.  Disabling plugin")
            self.enabled = False
        self.enabled = True

    @trace.trace
    def process(self, packet: packets.MessagePacket):
        LOG.info("Query COMMAND")

        fromcall = packet.from_call
        message = packet.get("message_text", None)

        pkt_tracker = tracker.PacketTrack()
        now = datetime.datetime.now()
        reply = "Pending messages ({}) {}".format(
            len(pkt_tracker),
            now.strftime("%H:%M:%S"),
        )

        searchstring = "^" + CONF.query_plugin.callsign + ".*"
        # only I can do admin commands
        if re.search(searchstring, fromcall):

            # resend last N most recent:  "!3"
            r = re.search(r"^\!([0-9]).*", message)
            if r is not None:
                if len(pkt_tracker) > 0:
                    last_n = r.group(1)
                    reply = packets.NULL_MESSAGE
                    LOG.debug(reply)
                    pkt_tracker.restart_delayed(count=int(last_n))
                else:
                    reply = "No pending msgs to resend"
                    LOG.debug(reply)
                return reply

            # resend all:   "!a"
            r = re.search(r"^\![aA].*", message)
            if r is not None:
                if len(pkt_tracker) > 0:
                    reply = packets.NULL_MESSAGE
                    LOG.debug(reply)
                    pkt_tracker.restart_delayed()
                else:
                    reply = "No pending msgs"
                    LOG.debug(reply)
                return reply

            # delete all:   "!d"
            r = re.search(r"^\![dD].*", message)
            if r is not None:
                reply = "Deleted ALL pending msgs."
                LOG.debug(reply)
                pkt_tracker.flush()
                return reply

        return reply



import fnmatch
import importlib
import inspect
import logging
import os
import pkgutil
import re
import sys
from traceback import print_tb
from urllib.parse import urljoin

from bs4 import BeautifulSoup
import click
import requests
from rich.console import Console
from rich.table import Table
from rich.text import Text
from thesmuggler import smuggle

from aprsd import cli_helper
from aprsd import plugin as aprsd_plugin
from aprsd.main import cli
from aprsd.plugins import (
    email, fortune, location, notify, ping, query, time, version, weather,
)


LOG = logging.getLogger("APRSD")
PYPI_URL = "https://pypi.org/search/"


def onerror(name):
    print(f"Error importing module {name}")
    type, value, traceback = sys.exc_info()
    print_tb(traceback)


def is_plugin(obj):
    for c in inspect.getmro(obj):
        if issubclass(c, aprsd_plugin.APRSDPluginBase):
            return True

    return False


def plugin_type(obj):
    for c in inspect.getmro(obj):
        if issubclass(c, aprsd_plugin.APRSDRegexCommandPluginBase):
            return "RegexCommand"
        if issubclass(c, aprsd_plugin.APRSDWatchListPluginBase):
            return "WatchList"
        if issubclass(c, aprsd_plugin.APRSDPluginBase):
            return "APRSDPluginBase"

    return "Unknown"


def walk_package(package):
    return pkgutil.walk_packages(
        package.__path__,
        package.__name__ + ".",
        onerror=onerror,
    )


def get_module_info(package_name, module_name, module_path):
    if not os.path.exists(module_path):
        return None

    dir_path = os.path.realpath(module_path)
    pattern = "*.py"

    obj_list = []

    for path, _subdirs, files in os.walk(dir_path):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                module = smuggle(f"{path}/{name}")
                for mem_name, obj in inspect.getmembers(module):
                    if inspect.isclass(obj) and is_plugin(obj):
                        obj_list.append(
                            {
                                "package": package_name,
                                "name": mem_name, "obj": obj,
                                "version": obj.version,
                                "path": f"{'.'.join([module_name, obj.__name__])}",
                            },
                        )

    return obj_list


def _get_installed_aprsd_items():
    # installed plugins
    plugins = {}
    extensions = {}
    for finder, name, ispkg in pkgutil.iter_modules():
        if name.startswith("aprsd_"):
            print(f"Found aprsd_ module: {name}")
            if ispkg:
                module = importlib.import_module(name)
                pkgs = walk_package(module)
                for pkg in pkgs:
                    pkg_info = get_module_info(module.__name__, pkg.name, module.__path__[0])
                    if "plugin" in name:
                        plugins[name] = pkg_info
                    elif "extension" in name:
                        extensions[name] = pkg_info
    return plugins, extensions


def get_installed_plugins():
    # installed plugins
    plugins, extensions = _get_installed_aprsd_items()
    return plugins


def get_installed_extensions():
    # installed plugins
    plugins, extensions = _get_installed_aprsd_items()
    return extensions


def show_built_in_plugins(console):
    modules = [email, fortune, location, notify, ping, query, time, version, weather]
    plugins = []

    for module in modules:
        entries = inspect.getmembers(module, inspect.isclass)
        for entry in entries:
            cls = entry[1]
            if issubclass(cls, aprsd_plugin.APRSDPluginBase):
                info = {
                    "name": cls.__qualname__,
                    "path": f"{cls.__module__}.{cls.__qualname__}",
                    "version": cls.version,
                    "docstring": cls.__doc__,
                    "short_desc": cls.short_description,
                }

                if issubclass(cls, aprsd_plugin.APRSDRegexCommandPluginBase):
                    info["command_regex"] = cls.command_regex
                    info["type"] = "RegexCommand"

                if issubclass(cls, aprsd_plugin.APRSDWatchListPluginBase):
                    info["type"] = "WatchList"

                plugins.append(info)

    plugins = sorted(plugins, key=lambda i: i["name"])

    table = Table(
        title="[not italic]:snake:[/] [bold][magenta]APRSD Built-in Plugins [not italic]:snake:[/]",
    )
    table.add_column("Plugin Name", style="cyan", no_wrap=True)
    table.add_column("Info", style="bold yellow")
    table.add_column("Type", style="bold green")
    table.add_column("Plugin Path", style="bold blue")
    for entry in plugins:
        table.add_row(entry["name"], entry["short_desc"], entry["type"], entry["path"])

    console.print(table)


def _get_pypi_packages():
    query = "aprsd"
    snippets = []
    s = requests.Session()
    for page in range(1, 3):
        params = {"q": query, "page": page}
        r = s.get(PYPI_URL, params=params)
        soup = BeautifulSoup(r.text, "html.parser")
        snippets += soup.select('a[class*="snippet"]')
        if not hasattr(s, "start_url"):
            s.start_url = r.url.rsplit("&page", maxsplit=1).pop(0)

    return snippets


def show_pypi_plugins(installed_plugins, console):
    snippets = _get_pypi_packages()

    title = Text.assemble(
        ("Pypi.org APRSD Installable Plugin Packages\n\n", "bold magenta"),
        ("Install any of the following plugins with\n", "bold yellow"),
        ("'pip install ", "bold white"),
        ("<Plugin Package Name>'", "cyan"),
    )

    table = Table(title=title)
    table.add_column("Plugin Package Name", style="cyan", no_wrap=True)
    table.add_column("Description", style="yellow")
    table.add_column("Version", style="yellow", justify="center")
    table.add_column("Released", style="bold green", justify="center")
    table.add_column("Installed?", style="red", justify="center")
    for snippet in snippets:
        link = urljoin(PYPI_URL, snippet.get("href"))
        package = re.sub(r"\s+", " ", snippet.select_one('span[class*="name"]').text.strip())
        version = re.sub(r"\s+", " ", snippet.select_one('span[class*="version"]').text.strip())
        created = re.sub(r"\s+", " ", snippet.select_one('span[class*="created"]').text.strip())
        description = re.sub(r"\s+", " ", snippet.select_one('p[class*="description"]').text.strip())
        emoji = ":open_file_folder:"

        if "aprsd-" not in package or "-plugin" not in package:
            continue

        under = package.replace("-", "_")
        if under in installed_plugins:
            installed = "Yes"
        else:
            installed = "No"

        table.add_row(
            f"[link={link}]{emoji}[/link] {package}",
            description, version, created, installed,
        )

    console.print("\n")
    console.print(table)


def show_pypi_extensions(installed_extensions, console):
    snippets = _get_pypi_packages()

    title = Text.assemble(
        ("Pypi.org APRSD Installable Extension Packages\n\n", "bold magenta"),
        ("Install any of the following extensions by running\n", "bold yellow"),
        ("'pip install ", "bold white"),
        ("<Plugin Package Name>'", "cyan"),
    )
    table = Table(title=title)
    table.add_column("Extension Package Name", style="cyan", no_wrap=True)
    table.add_column("Description", style="yellow")
    table.add_column("Version", style="yellow", justify="center")
    table.add_column("Released", style="bold green", justify="center")
    table.add_column("Installed?", style="red", justify="center")
    for snippet in snippets:
        link = urljoin(PYPI_URL, snippet.get("href"))
        package = re.sub(r"\s+", " ", snippet.select_one('span[class*="name"]').text.strip())
        version = re.sub(r"\s+", " ", snippet.select_one('span[class*="version"]').text.strip())
        created = re.sub(r"\s+", " ", snippet.select_one('span[class*="created"]').text.strip())
        description = re.sub(r"\s+", " ", snippet.select_one('p[class*="description"]').text.strip())
        emoji = ":open_file_folder:"

        if "aprsd-" not in package or "-extension" not in package:
            continue

        under = package.replace("-", "_")
        if under in installed_extensions:
            installed = "Yes"
        else:
            installed = "No"

        table.add_row(
            f"[link={link}]{emoji}[/link] {package}",
            description, version, created, installed,
        )

    console.print("\n")
    console.print(table)


def show_installed_plugins(installed_plugins, console):
    if not installed_plugins:
        return

    table = Table(
        title="[not italic]:snake:[/] [bold][magenta]APRSD Installed 3rd party Plugins [not italic]:snake:[/]",
    )
    table.add_column("Package Name", style=" bold white", no_wrap=True)
    table.add_column("Plugin Name", style="cyan", no_wrap=True)
    table.add_column("Version", style="yellow", justify="center")
    table.add_column("Type", style="bold green")
    table.add_column("Plugin Path", style="bold blue")
    for name in installed_plugins:
        for plugin in installed_plugins[name]:
            table.add_row(
                name.replace("_", "-"),
                plugin["name"],
                plugin["version"],
                plugin_type(plugin["obj"]),
                plugin["path"],
            )

    console.print("\n")
    console.print(table)


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.pass_context
@cli_helper.process_standard_options_no_config
def list_plugins(ctx):
    """List the built in plugins available to APRSD."""
    console = Console()

    with console.status("Show Built-in Plugins") as status:
        show_built_in_plugins(console)

        status.update("Fetching pypi.org plugins")
        installed_plugins = get_installed_plugins()
        show_pypi_plugins(installed_plugins, console)

        status.update("Looking for installed APRSD plugins")
        show_installed_plugins(installed_plugins, console)


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.pass_context
@cli_helper.process_standard_options_no_config
def list_extensions(ctx):
    """List the built in plugins available to APRSD."""
    console = Console()

    with console.status("Show APRSD Extensions") as status:
        status.update("Fetching pypi.org APRSD Extensions")
        installed_extensions = get_installed_extensions()
        show_pypi_extensions(installed_extensions, console)



import abc
import datetime
import logging
import threading

import wrapt


LOG = logging.getLogger("APRSD")


class APRSDThreadList:
    """Singleton class that keeps track of application wide threads."""

    _instance = None

    threads_list = []
    lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls.threads_list = []
        return cls._instance

    @wrapt.synchronized(lock)
    def add(self, thread_obj):
        self.threads_list.append(thread_obj)

    @wrapt.synchronized(lock)
    def remove(self, thread_obj):
        self.threads_list.remove(thread_obj)

    @wrapt.synchronized(lock)
    def stop_all(self):
        """Iterate over all threads and call stop on them."""
        for th in self.threads_list:
            LOG.info(f"Stopping Thread {th.name}")
            if hasattr(th, "packet"):
                LOG.info(F"{th.name} packet {th.packet}")
            th.stop()

    @wrapt.synchronized(lock)
    def __len__(self):
        return len(self.threads_list)


class APRSDThread(threading.Thread, metaclass=abc.ABCMeta):

    def __init__(self, name):
        super().__init__(name=name)
        self.thread_stop = False
        APRSDThreadList().add(self)
        self._last_loop = datetime.datetime.now()

    def _should_quit(self):
        """ see if we have a quit message from the global queue."""
        if self.thread_stop:
            return True

    def stop(self):
        self.thread_stop = True

    @abc.abstractmethod
    def loop(self):
        pass

    def _cleanup(self):
        """Add code to subclass to do any cleanup"""

    def __str__(self):
        out = f"Thread <{self.__class__.__name__}({self.name}) Alive? {self.is_alive()}>"
        return out

    def loop_age(self):
        """How old is the last loop call?"""
        return datetime.datetime.now() - self._last_loop

    def run(self):
        LOG.debug("Starting")
        while not self._should_quit():
            can_loop = self.loop()
            self._last_loop = datetime.datetime.now()
            if not can_loop:
                self.stop()
        self._cleanup()
        APRSDThreadList().remove(self)
        LOG.debug("Exiting")



#
# License GPLv2
#

# python included libs
import datetime
import logging
import signal
import sys
import time

import click
from oslo_config import cfg
from rich.console import Console

# local imports here
import aprsd
from aprsd import cli_helper, client, packets, plugin, stats, threads
from aprsd.main import cli
from aprsd.rpc import server as rpc_server
from aprsd.threads import rx


# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
LOG = logging.getLogger("APRSD")
CONF = cfg.CONF
console = Console()


def signal_handler(sig, frame):
    threads.APRSDThreadList().stop_all()
    if "subprocess" not in str(frame):
        LOG.info(
            "Ctrl+C, Sending all threads exit! Can take up to 10 seconds {}".format(
                datetime.datetime.now(),
            ),
        )
        time.sleep(5)
        LOG.info(stats.APRSDStats())


class APRSDListenThread(rx.APRSDRXThread):
    def __init__(self, packet_queue, packet_filter=None, plugin_manager=None):
        super().__init__(packet_queue)
        self.packet_filter = packet_filter
        self.plugin_manager = plugin_manager
        if self.plugin_manager:
            LOG.info(f"Plugins {self.plugin_manager.get_message_plugins()}")

    def process_packet(self, *args, **kwargs):
        packet = self._client.decode_packet(*args, **kwargs)
        filters = {
            packets.Packet.__name__: packets.Packet,
            packets.AckPacket.__name__: packets.AckPacket,
            packets.GPSPacket.__name__: packets.GPSPacket,
            packets.MessagePacket.__name__: packets.MessagePacket,
            packets.MicEPacket.__name__: packets.MicEPacket,
            packets.WeatherPacket.__name__: packets.WeatherPacket,
        }

        if self.packet_filter:
            filter_class = filters[self.packet_filter]
            if isinstance(packet, filter_class):
                packet.log(header="RX")
                if self.plugin_manager:
                    # Don't do anything with the reply
                    # This is the listen only command.
                    self.plugin_manager.run(packet)
        else:
            if self.plugin_manager:
                # Don't do anything with the reply.
                # This is the listen only command.
                self.plugin_manager.run(packet)
            else:
                packet.log(header="RX")

        packets.PacketList().rx(packet)


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "--aprs-login",
    envvar="APRS_LOGIN",
    show_envvar=True,
    help="What callsign to send the message from.",
)
@click.option(
    "--aprs-password",
    envvar="APRS_PASSWORD",
    show_envvar=True,
    help="the APRS-IS password for APRS_LOGIN",
)
@click.option(
    "--packet-filter",
    type=click.Choice(
        [
            packets.Packet.__name__,
            packets.AckPacket.__name__,
            packets.GPSPacket.__name__,
            packets.MicEPacket.__name__,
            packets.MessagePacket.__name__,
            packets.WeatherPacket.__name__,
        ],
        case_sensitive=False,
    ),
    help="Filter by packet type",
)
@click.option(
    "--load-plugins",
    default=False,
    is_flag=True,
    help="Load plugins as enabled in aprsd.conf ?",
)
@click.argument(
    "filter",
    nargs=-1,
    required=True,
)
@click.pass_context
@cli_helper.process_standard_options
def listen(
    ctx,
    aprs_login,
    aprs_password,
    packet_filter,
    load_plugins,
    filter,
):
    """Listen to packets on the APRS-IS Network based on FILTER.

    FILTER is the APRS Filter to use.\n
     see http://www.aprs-is.net/javAPRSFilter.aspx\n
    r/lat/lon/dist - Range Filter Pass posits and objects within dist km from lat/lon.\n
    p/aa/bb/cc... - Prefix Filter Pass traffic with fromCall that start with aa or bb or cc.\n
    b/call1/call2... - Budlist Filter Pass all traffic from exact call: call1, call2, ... (* wild card allowed) \n
    o/obj1/obj2... - Object Filter Pass all objects with the exact name of obj1, obj2, ... (* wild card allowed)\n

    """
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    if not aprs_login:
        click.echo(ctx.get_help())
        click.echo("")
        ctx.fail("Must set --aprs-login or APRS_LOGIN")
        ctx.exit()

    if not aprs_password:
        click.echo(ctx.get_help())
        click.echo("")
        ctx.fail("Must set --aprs-password or APRS_PASSWORD")
        ctx.exit()

    # CONF.aprs_network.login = aprs_login
    # config["aprs"]["password"] = aprs_password

    LOG.info(f"APRSD Listen Started version: {aprsd.__version__}")

    CONF.log_opt_values(LOG, logging.DEBUG)

    # Try and load saved MsgTrack list
    LOG.debug("Loading saved MsgTrack object.")

    # Initialize the client factory and create
    # The correct client object ready for use
    client.ClientFactory.setup()
    # Make sure we have 1 client transport enabled
    if not client.factory.is_client_enabled():
        LOG.error("No Clients are enabled in config.")
        sys.exit(-1)

    # Creates the client object
    LOG.info("Creating client connection")
    aprs_client = client.factory.create()
    LOG.info(aprs_client)

    LOG.debug(f"Filter by '{filter}'")
    aprs_client.set_filter(filter)

    keepalive = threads.KeepAliveThread()
    keepalive.start()

    if CONF.rpc_settings.enabled:
        rpc = rpc_server.APRSDRPCThread()
        rpc.start()

    pm = None
    pm = plugin.PluginManager()
    if load_plugins:
        LOG.info("Loading plugins")
        pm.setup_plugins(load_help_plugin=False)
    else:
        LOG.warning(
            "Not Loading any plugins use --load-plugins to load what's "
            "defined in the config file.",
        )

    LOG.debug("Create APRSDListenThread")
    listen_thread = APRSDListenThread(
        packet_queue=threads.packet_queue,
        packet_filter=packet_filter,
        plugin_manager=pm,
    )
    LOG.debug("Start APRSDListenThread")
    listen_thread.start()
    LOG.debug("keepalive Join")
    keepalive.join()
    LOG.debug("listen_thread Join")
    listen_thread.join()

    if CONF.rpc_settings.enabled:
        rpc.join()



import rpyc


class AuthSocketStream(rpyc.SocketStream):
    """Used to authenitcate the RPC stream to remote."""

    @classmethod
    def connect(cls, *args, authorizer=None, **kwargs):
        stream_obj = super().connect(*args, **kwargs)

        if callable(authorizer):
            authorizer(stream_obj.sock)

        return stream_obj



from unittest import mock

from oslo_config import cfg

from aprsd import conf  # noqa: F401
from aprsd.plugins import location as location_plugin

from .. import fake, test_plugin


CONF = cfg.CONF


class TestLocationPlugin(test_plugin.TestPlugin):

    def test_location_not_enabled_missing_aprs_fi_key(self):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        CONF.aprs_fi.apiKey = None
        fortune = location_plugin.LocationPlugin()
        expected = "LocationPlugin isn't enabled"
        packet = fake.fake_packet(message="location")
        actual = fortune.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    def test_location_failed_aprs_fi_location(self, mock_check):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        mock_check.side_effect = Exception
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        fortune = location_plugin.LocationPlugin()
        expected = "Failed to fetch aprs.fi location"
        packet = fake.fake_packet(message="location")
        actual = fortune.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    def test_location_failed_aprs_fi_location_no_entries(self, mock_check):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        mock_check.return_value = {"entries": []}
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        fortune = location_plugin.LocationPlugin()
        expected = "Failed to fetch aprs.fi location"
        packet = fake.fake_packet(message="location")
        actual = fortune.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    @mock.patch("geopy.geocoders.Nominatim.reverse")
    @mock.patch("time.time")
    def test_location_unknown_gps(self, mock_time, mock_geocode, mock_check_aprs):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        mock_check_aprs.return_value = {
            "entries": [
                {
                    "lat": 1,
                    "lng": 1,
                    "lasttime": 10,
                },
            ],
        }
        mock_geocode.side_effect = Exception
        mock_time.return_value = 10
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        fortune = location_plugin.LocationPlugin()
        expected = "KFAKE: Unknown Location 0' 1.00,1.00 0.0h ago"
        packet = fake.fake_packet(message="location")
        actual = fortune.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    @mock.patch("geopy.geocoders.Nominatim.reverse")
    @mock.patch("time.time")
    def test_location_works(self, mock_time, mock_geocode, mock_check_aprs):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        mock_check_aprs.return_value = {
            "entries": [
                {
                    "lat": 1,
                    "lng": 1,
                    "lasttime": 10,
                },
            ],
        }
        expected = "Appomattox"
        state = "VA"

        class TempLocation:
            raw = {
                "address": {
                    "county": expected,
                    "country_code": "us",
                    "state": state,
                    "country": "United States",
                },
            }
        mock_geocode.return_value = TempLocation()
        mock_time.return_value = 10
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        fortune = location_plugin.LocationPlugin()
        expected = f"KFAKE: {expected}, {state} 0' 1.00,1.00 0.0h ago"
        packet = fake.fake_packet(message="location")
        actual = fortune.filter(packet)
        self.assertEqual(expected, actual)



import datetime
import logging
import time
import tracemalloc

from oslo_config import cfg

from aprsd import client, packets, stats, utils
from aprsd.threads import APRSDThread, APRSDThreadList


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class KeepAliveThread(APRSDThread):
    cntr = 0
    checker_time = datetime.datetime.now()

    def __init__(self):
        tracemalloc.start()
        super().__init__("KeepAlive")
        max_timeout = {"hours": 0.0, "minutes": 2, "seconds": 0}
        self.max_delta = datetime.timedelta(**max_timeout)

    def loop(self):
        if self.cntr % 60 == 0:
            pkt_tracker = packets.PacketTrack()
            stats_obj = stats.APRSDStats()
            pl = packets.PacketList()
            thread_list = APRSDThreadList()
            now = datetime.datetime.now()
            last_email = stats_obj.email_thread_time
            if last_email:
                email_thread_time = utils.strfdelta(now - last_email)
            else:
                email_thread_time = "N/A"

            last_msg_time = utils.strfdelta(now - stats_obj.aprsis_keepalive)

            current, peak = tracemalloc.get_traced_memory()
            stats_obj.set_memory(current)
            stats_obj.set_memory_peak(peak)

            login = CONF.callsign

            tracked_packets = len(pkt_tracker)

            keepalive = (
                "{} - Uptime {} RX:{} TX:{} Tracker:{} Msgs TX:{} RX:{} "
                "Last:{} Email: {} - RAM Current:{} Peak:{} Threads:{}"
            ).format(
                login,
                utils.strfdelta(stats_obj.uptime),
                pl.total_rx(),
                pl.total_tx(),
                tracked_packets,
                stats_obj._pkt_cnt["MessagePacket"]["tx"],
                stats_obj._pkt_cnt["MessagePacket"]["rx"],
                last_msg_time,
                email_thread_time,
                utils.human_size(current),
                utils.human_size(peak),
                len(thread_list),
            )
            LOG.info(keepalive)
            thread_out = []
            thread_info = {}
            for thread in thread_list.threads_list:
                alive = thread.is_alive()
                age = thread.loop_age()
                key = thread.__class__.__name__
                thread_out.append(f"{key}:{alive}:{age}")
                if key not in thread_info:
                    thread_info[key] = {}
                thread_info[key]["alive"] = alive
                thread_info[key]["age"] = age
                if not alive:
                    LOG.error(f"Thread {thread}")
            LOG.info(",".join(thread_out))
            stats_obj.set_thread_info(thread_info)

            # check the APRS connection
            cl = client.factory.create()
            # Reset the connection if it's dead and this isn't our
            # First time through the loop.
            # The first time through the loop can happen at startup where
            # The keepalive thread starts before the client has a chance
            # to make it's connection the first time.
            if not cl.is_alive() and self.cntr > 0:
                LOG.error(f"{cl.__class__.__name__} is not alive!!! Resetting")
                client.factory.create().reset()
            else:
                # See if we should reset the aprs-is client
                # Due to losing a keepalive from them
                delta_dict = utils.parse_delta_str(last_msg_time)
                delta = datetime.timedelta(**delta_dict)

                if delta > self.max_delta:
                    #  We haven't gotten a keepalive from aprs-is in a while
                    # reset the connection.a
                    if not client.KISSClient.is_enabled():
                        LOG.warning(f"Resetting connection to APRS-IS {delta}")
                        client.factory.create().reset()

            # Check version every day
            delta = now - self.checker_time
            if delta > datetime.timedelta(hours=24):
                self.checker_time = now
                level, msg = utils._check_version()
                if level:
                    LOG.warning(msg)
        self.cntr += 1
        time.sleep(1)
        return True



"""
The options for log setup
"""
import logging

from oslo_config import cfg


LOG_LEVELS = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
}

DEFAULT_DATE_FORMAT = "%m/%d/%Y %I:%M:%S %p"
DEFAULT_LOG_FORMAT = (
    "[%(asctime)s] [%(threadName)-20.20s] [%(levelname)-5.5s]"
    " %(message)s - [%(pathname)s:%(lineno)d]"
)

DEFAULT_LOG_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
    "<yellow>{thread.name: <18}</yellow> | "
    "<level>{level: <8}</level> | "
    "<level>{message}</level> | "
    "<cyan>{name}</cyan>:<cyan>{function:}</cyan>:<magenta>{line:}</magenta>"
)

logging_group = cfg.OptGroup(
    name="logging",
    title="Logging options",
)
logging_opts = [
    cfg.StrOpt(
        "logfile",
        default=None,
        help="File to log to",
    ),
    cfg.StrOpt(
        "logformat",
        default=DEFAULT_LOG_FORMAT,
        help="Log file format, unless rich_logging enabled.",
    ),
    cfg.StrOpt(
        "log_level",
        default="INFO",
        choices=LOG_LEVELS.keys(),
        help="Log level for logging of events.",
    ),
]


def register_opts(config):
    config.register_group(logging_group)
    config.register_opts(logging_opts, group=logging_group)


def list_opts():
    return {
        logging_group.name: (
            logging_opts
        ),
    }



# Fetch active stats from a remote running instance of aprsd server
# This uses the RPC server to fetch the stats from the remote server.

import logging

import click
from oslo_config import cfg
from rich.console import Console
from rich.table import Table

# local imports here
import aprsd
from aprsd import cli_helper
from aprsd.main import cli
from aprsd.rpc import client as rpc_client


# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
LOG = logging.getLogger("APRSD")
CONF = cfg.CONF


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "--host", type=str,
    default=None,
    help="IP address of the remote aprsd server to fetch stats from.",
)
@click.option(
    "--port", type=int,
    default=None,
    help="Port of the remote aprsd server rpc port to fetch stats from.",
)
@click.option(
    "--magic-word", type=str,
    default=None,
    help="Magic word of the remote aprsd server rpc port to fetch stats from.",
)
@click.pass_context
@cli_helper.process_standard_options
def fetch_stats(ctx, host, port, magic_word):
    """Fetch stats from a remote running instance of aprsd server."""
    LOG.info(f"APRSD Fetch-Stats started version: {aprsd.__version__}")

    CONF.log_opt_values(LOG, logging.DEBUG)
    if not host:
        host = CONF.rpc_settings.ip
    if not port:
        port = CONF.rpc_settings.port
    if not magic_word:
        magic_word = CONF.rpc_settings.magic_word

    msg = f"Fetching stats from {host}:{port} with magic word '{magic_word}'"
    console = Console()
    console.print(msg)
    with console.status(msg):
        client = rpc_client.RPCClient(host, port, magic_word)
        stats = client.get_stats_dict()
        if stats:
            console.print_json(data=stats)
        else:
            LOG.error(f"Failed to fetch stats via RPC aprsd server at {host}:{port}")
            return
    aprsd_title = (
        "APRSD "
        f"[bold cyan]v{stats['aprsd']['version']}[/] "
        f"Callsign [bold green]{stats['aprsd']['callsign']}[/] "
        f"Uptime [bold yellow]{stats['aprsd']['uptime']}[/]"
    )

    console.rule(f"Stats from {host}:{port} with magic word '{magic_word}'")
    console.print("\n\n")
    console.rule(aprsd_title)

    # Show the connection to APRS
    # It can be a connection to an APRS-IS server or a local TNC via KISS or KISSTCP
    if "aprs-is" in stats:
        title = f"APRS-IS Connection {stats['aprs-is']['server']}"
        table = Table(title=title)
        table.add_column("Key")
        table.add_column("Value")
        for key, value in stats["aprs-is"].items():
            table.add_row(key, value)
        console.print(table)

    threads_table = Table(title="Threads")
    threads_table.add_column("Name")
    threads_table.add_column("Alive?")
    for name, alive in stats["aprsd"]["threads"].items():
        threads_table.add_row(name, str(alive))

    console.print(threads_table)

    msgs_table = Table(title="Messages")
    msgs_table.add_column("Key")
    msgs_table.add_column("Value")
    for key, value in stats["messages"].items():
        msgs_table.add_row(key, str(value))

    console.print(msgs_table)

    packet_totals = Table(title="Packet Totals")
    packet_totals.add_column("Key")
    packet_totals.add_column("Value")
    packet_totals.add_row("Total Received", str(stats["packets"]["total_received"]))
    packet_totals.add_row("Total Sent", str(stats["packets"]["total_sent"]))
    packet_totals.add_row("Total Tracked", str(stats["packets"]["total_tracked"]))
    console.print(packet_totals)

    # Show each of the packet types
    packets_table = Table(title="Packets By Type")
    packets_table.add_column("Packet Type")
    packets_table.add_column("TX")
    packets_table.add_column("RX")
    for key, value in stats["packets"]["by_type"].items():
        packets_table.add_row(key, str(value["tx"]), str(value["rx"]))

    console.print(packets_table)

    if "plugins" in stats:
        count = len(stats["plugins"])
        plugins_table = Table(title=f"Plugins ({count})")
        plugins_table.add_column("Plugin")
        plugins_table.add_column("Enabled")
        plugins_table.add_column("Version")
        plugins_table.add_column("TX")
        plugins_table.add_column("RX")
        for key, value in stats["plugins"].items():
            plugins_table.add_row(
                key,
                str(stats["plugins"][key]["enabled"]),
                stats["plugins"][key]["version"],
                str(stats["plugins"][key]["tx"]),
                str(stats["plugins"][key]["rx"]),
            )

        console.print(plugins_table)

    if "seen_list" in stats["aprsd"]:
        count = len(stats["aprsd"]["seen_list"])
        seen_table = Table(title=f"Seen List ({count})")
        seen_table.add_column("Callsign")
        seen_table.add_column("Message Count")
        seen_table.add_column("Last Heard")
        for key, value in stats["aprsd"]["seen_list"].items():
            seen_table.add_row(key, str(value["count"]), value["last"])

        console.print(seen_table)

    if "watch_list" in stats["aprsd"]:
        count = len(stats["aprsd"]["watch_list"])
        watch_table = Table(title=f"Watch List ({count})")
        watch_table.add_column("Callsign")
        watch_table.add_column("Last Heard")
        for key, value in stats["aprsd"]["watch_list"].items():
            watch_table.add_row(key, value["last"])

        console.print(watch_table)



from aprsd.packets.core import (  # noqa: F401
    AckPacket, BeaconPacket, GPSPacket, MessagePacket, MicEPacket, Packet,
    RejectPacket, StatusPacket, WeatherPacket,
)
from aprsd.packets.packet_list import PacketList  # noqa: F401
from aprsd.packets.seen_list import SeenList  # noqa: F401
from aprsd.packets.tracker import PacketTrack  # noqa: F401
from aprsd.packets.watch_list import WatchList  # noqa: F401


NULL_MESSAGE = -1



import sys
import unittest

from aprsd.plugins import email


if sys.version_info >= (3, 2):
    from unittest import mock
else:
    from unittest import mock


class TestMain(unittest.TestCase):
    @mock.patch("aprsd.plugins.email._imap_connect")
    @mock.patch("aprsd.plugins.email._smtp_connect")
    def test_validate_email(self, imap_mock, smtp_mock):
        """Test to make sure we fail."""
        imap_mock.return_value = None
        smtp_mock.return_value = {"smaiof": "fire"}
        mock.MagicMock()

        email.validate_email_config(True)



import datetime
import logging
import threading

from oslo_config import cfg
import wrapt

from aprsd import utils
from aprsd.utils import objectstore


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class WatchList(objectstore.ObjectStoreMixin):
    """Global watch list and info for callsigns."""

    _instance = None
    lock = threading.Lock()
    data = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._init_store()
            cls._instance.data = {}
        return cls._instance

    def __init__(self, config=None):
        ring_size = CONF.watch_list.packet_keep_count

        if CONF.watch_list.callsigns:
            for callsign in CONF.watch_list.callsigns:
                call = callsign.replace("*", "")
                # FIXME(waboring) - we should fetch the last time we saw
                # a beacon from a callsign or some other mechanism to find
                # last time a message was seen by aprs-is.  For now this
                # is all we can do.
                self.data[call] = {
                    "last": datetime.datetime.now(),
                    "packets": utils.RingBuffer(
                        ring_size,
                    ),
                }

    def is_enabled(self):
        return CONF.watch_list.enabled

    def callsign_in_watchlist(self, callsign):
        return callsign in self.data

    @wrapt.synchronized(lock)
    def update_seen(self, packet):
        if packet.addresse:
            callsign = packet.addresse
        else:
            callsign = packet.from_call
        if self.callsign_in_watchlist(callsign):
            self.data[callsign]["last"] = datetime.datetime.now()
            self.data[callsign]["packets"].append(packet)

    def last_seen(self, callsign):
        if self.callsign_in_watchlist(callsign):
            return self.data[callsign]["last"]

    def age(self, callsign):
        now = datetime.datetime.now()
        return str(now - self.last_seen(callsign))

    def max_delta(self, seconds=None):
        if not seconds:
            seconds = CONF.watch_list.alert_time_seconds
        max_timeout = {"seconds": seconds}
        return datetime.timedelta(**max_timeout)

    def is_old(self, callsign, seconds=None):
        """Watch list callsign last seen is old compared to now?

        This tests to see if the last time we saw a callsign packet,
        if that is older than the allowed timeout in the config.

        We put this here so any notification plugin can use this
        same test.
        """
        age = self.age(callsign)

        delta = utils.parse_delta_str(age)
        d = datetime.timedelta(**delta)

        max_delta = self.max_delta(seconds=seconds)

        if d > max_delta:
            return True
        else:
            return False



import click
import click_completion

from aprsd.main import cli


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@cli.group(help="Click Completion subcommands", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def completion(ctx):
    pass


# show dumps out the completion code for a particular shell
@completion.command(help="Show completion code for shell", name="show")
@click.option("-i", "--case-insensitive/--no-case-insensitive", help="Case insensitive completion")
@click.argument("shell", required=False, type=click_completion.DocumentedChoice(click_completion.core.shells))
def show(shell, case_insensitive):
    """Show the click-completion-command completion code"""
    extra_env = {"_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE": "ON"} if case_insensitive else {}
    click.echo(click_completion.core.get_code(shell, extra_env=extra_env))


# install will install the completion code for a particular shell
@completion.command(help="Install completion code for a shell", name="install")
@click.option("--append/--overwrite", help="Append the completion code to the file", default=None)
@click.option("-i", "--case-insensitive/--no-case-insensitive", help="Case insensitive completion")
@click.argument("shell", required=False, type=click_completion.DocumentedChoice(click_completion.core.shells))
@click.argument("path", required=False)
def install(append, case_insensitive, shell, path):
    """Install the click-completion-command completion"""
    extra_env = {"_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE": "ON"} if case_insensitive else {}
    shell, path = click_completion.core.install(shell=shell, path=path, append=append, extra_env=extra_env)
    click.echo(f"{shell} completion installed in {path}")



import datetime
import threading

from oslo_config import cfg
import wrapt

from aprsd.threads import tx
from aprsd.utils import objectstore


CONF = cfg.CONF


class PacketTrack(objectstore.ObjectStoreMixin):
    """Class to keep track of outstanding text messages.

    This is a thread safe class that keeps track of active
    messages.

    When a message is asked to be sent, it is placed into this
    class via it's id.  The TextMessage class's send() method
    automatically adds itself to this class.  When the ack is
    recieved from the radio, the message object is removed from
    this class.
    """

    _instance = None
    _start_time = None
    lock = threading.Lock()

    data: dict = {}
    total_tracked: int = 0

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._start_time = datetime.datetime.now()
            cls._instance._init_store()
        return cls._instance

    @wrapt.synchronized(lock)
    def __getitem__(self, name):
        return self.data[name]

    @wrapt.synchronized(lock)
    def __iter__(self):
        return iter(self.data)

    @wrapt.synchronized(lock)
    def keys(self):
        return self.data.keys()

    @wrapt.synchronized(lock)
    def items(self):
        return self.data.items()

    @wrapt.synchronized(lock)
    def values(self):
        return self.data.values()

    @wrapt.synchronized(lock)
    def __len__(self):
        return len(self.data)

    @wrapt.synchronized(lock)
    def add(self, packet):
        key = packet.msgNo
        packet._last_send_attempt = 0
        self.data[key] = packet
        self.total_tracked += 1

    @wrapt.synchronized(lock)
    def get(self, key):
        return self.data.get(key, None)

    @wrapt.synchronized(lock)
    def remove(self, key):
        try:
            del self.data[key]
        except KeyError:
            pass

    def restart(self):
        """Walk the list of messages and restart them if any."""
        for key in self.data.keys():
            pkt = self.data[key]
            if pkt._last_send_attempt < pkt.retry_count:
                tx.send(pkt)

    def _resend(self, packet):
        packet._last_send_attempt = 0
        tx.send(packet)

    def restart_delayed(self, count=None, most_recent=True):
        """Walk the list of delayed messages and restart them if any."""
        if not count:
            # Send all the delayed messages
            for key in self.data.keys():
                pkt = self.data[key]
                if pkt._last_send_attempt == pkt._retry_count:
                    self._resend(pkt)
        else:
            # They want to resend <count> delayed messages
            tmp = sorted(
                self.data.items(),
                reverse=most_recent,
                key=lambda x: x[1].last_send_time,
            )
            pkt_list = tmp[:count]
            for (_key, pkt) in pkt_list:
                self._resend(pkt)



#
#  Dev.py is used to help develop plugins
#
#
# python included libs
import logging

import click
from oslo_config import cfg

# local imports here
from aprsd import cli_helper, client, conf, packets, plugin
from aprsd.main import cli
from aprsd.utils import trace


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@cli.group(help="Development type subcommands", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def dev(ctx):
    pass


@dev.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "--aprs-login",
    envvar="APRS_LOGIN",
    show_envvar=True,
    help="What callsign to send the message from.",
)
@click.option(
    "-p",
    "--plugin",
    "plugin_path",
    show_default=True,
    default=None,
    help="The plugin to run.  Ex: aprsd.plugins.ping.PingPlugin",
)
@click.option(
    "-a",
    "--all",
    "load_all",
    show_default=True,
    is_flag=True,
    default=False,
    help="Load all the plugins in config?",
)
@click.option(
    "-n",
    "--num",
    "number",
    show_default=True,
    default=1,
    help="Number of times to call the plugin",
)
@click.argument("message", nargs=-1, required=True)
@click.pass_context
@cli_helper.process_standard_options
def test_plugin(
    ctx,
    aprs_login,
    plugin_path,
    load_all,
    number,
    message,
):
    """Test an individual APRSD plugin given a python path."""

    CONF.log_opt_values(LOG, logging.DEBUG)

    if not aprs_login:
        if CONF.aprs_network.login == conf.client.DEFAULT_LOGIN:
            click.echo("Must set --aprs_login or APRS_LOGIN")
            ctx.exit(-1)
            return
        else:
            fromcall = CONF.aprs_network.login
    else:
        fromcall = aprs_login

    if not plugin_path:
        click.echo(ctx.get_help())
        click.echo("")
        click.echo("Failed to provide -p option to test a plugin")
        ctx.exit(-1)
        return

    if type(message) is tuple:
        message = " ".join(message)

    if CONF.trace_enabled:
        trace.setup_tracing(["method", "api"])

    client.Client()

    pm = plugin.PluginManager()
    if load_all:
        pm.setup_plugins()
    obj = pm._create_class(plugin_path, plugin.APRSDPluginBase)
    if not obj:
        click.echo(ctx.get_help())
        click.echo("")
        ctx.fail(f"Failed to create object from plugin path '{plugin_path}'")
        ctx.exit()

    # Register the plugin they wanted tested.
    LOG.info(
        "Testing plugin {} Version {}".format(
            obj.__class__, obj.version,
        ),
    )
    pm.register_msg(obj)

    packet = packets.MessagePacket(
        from_call=fromcall,
        to_call=CONF.callsign,
        msgNo=1,
        message_text=message,
    )
    LOG.info(f"P'{plugin_path}'  F'{fromcall}'   C'{message}'")

    for x in range(number):
        replies = pm.run(packet)
        # Plugin might have threads, so lets stop them so we can exit.
        # obj.stop_threads()
        for reply in replies:
            if isinstance(reply, list):
                # one of the plugins wants to send multiple messages
                for subreply in reply:
                    if isinstance(subreply, packets.Packet):
                        LOG.info(subreply)
                    else:
                        LOG.info(
                            packets.MessagePacket(
                                from_call=CONF.callsign,
                                to_call=fromcall,
                                message_text=subreply,
                            ),
                        )
            elif isinstance(reply, packets.Packet):
                # We have a message based object.
                LOG.info(reply)
            else:
                # A plugin can return a null message flag which signals
                # us that they processed the message correctly, but have
                # nothing to reply with, so we avoid replying with a
                # usage string
                if reply is not packets.NULL_MESSAGE:
                    LOG.info(
                        packets.MessagePacket(
                            from_call=CONF.callsign,
                            to_call=fromcall,
                            message_text=reply,
                        ),
                    )
    pm.stop()



import datetime
import logging
import threading

from oslo_config import cfg
import wrapt

import aprsd
from aprsd import packets, plugin, utils


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class APRSDStats:

    _instance = None
    lock = threading.Lock()

    start_time = None
    _aprsis_server = None
    _aprsis_keepalive = None

    _email_thread_last_time = None
    _email_tx = 0
    _email_rx = 0

    _mem_current = 0
    _mem_peak = 0

    _thread_info = {}

    _pkt_cnt = {
        "Packet": {
            "tx": 0,
            "rx": 0,
        },
        "AckPacket": {
            "tx": 0,
            "rx": 0,
        },
        "GPSPacket": {
            "tx": 0,
            "rx": 0,
        },
        "StatusPacket": {
            "tx": 0,
            "rx": 0,
        },
        "MicEPacket": {
            "tx": 0,
            "rx": 0,
        },
        "MessagePacket": {
            "tx": 0,
            "rx": 0,
        },
        "WeatherPacket": {
            "tx": 0,
            "rx": 0,
        },
        "ObjectPacket": {
            "tx": 0,
            "rx": 0,
        },
    }

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # any init here
            cls._instance.start_time = datetime.datetime.now()
            cls._instance._aprsis_keepalive = datetime.datetime.now()
        return cls._instance

    @wrapt.synchronized(lock)
    @property
    def uptime(self):
        return datetime.datetime.now() - self.start_time

    @wrapt.synchronized(lock)
    @property
    def memory(self):
        return self._mem_current

    @wrapt.synchronized(lock)
    def set_memory(self, memory):
        self._mem_current = memory

    @wrapt.synchronized(lock)
    @property
    def memory_peak(self):
        return self._mem_peak

    @wrapt.synchronized(lock)
    def set_memory_peak(self, memory):
        self._mem_peak = memory

    @wrapt.synchronized(lock)
    def set_thread_info(self, thread_info):
        self._thread_info = thread_info

    @wrapt.synchronized(lock)
    @property
    def thread_info(self):
        return self._thread_info

    @wrapt.synchronized(lock)
    @property
    def aprsis_server(self):
        return self._aprsis_server

    @wrapt.synchronized(lock)
    def set_aprsis_server(self, server):
        self._aprsis_server = server

    @wrapt.synchronized(lock)
    @property
    def aprsis_keepalive(self):
        return self._aprsis_keepalive

    @wrapt.synchronized(lock)
    def set_aprsis_keepalive(self):
        self._aprsis_keepalive = datetime.datetime.now()

    def rx(self, packet):
        pkt_type = packet.__class__.__name__
        if pkt_type not in self._pkt_cnt:
            self._pkt_cnt[pkt_type] = {
                "tx": 0,
                "rx": 0,
            }
        self._pkt_cnt[pkt_type]["rx"] += 1

    def tx(self, packet):
        pkt_type = packet.__class__.__name__
        if pkt_type not in self._pkt_cnt:
            self._pkt_cnt[pkt_type] = {
                "tx": 0,
                "rx": 0,
            }
        self._pkt_cnt[pkt_type]["tx"] += 1

    @wrapt.synchronized(lock)
    @property
    def msgs_tracked(self):
        return packets.PacketTrack().total_tracked

    @wrapt.synchronized(lock)
    @property
    def email_tx(self):
        return self._email_tx

    @wrapt.synchronized(lock)
    def email_tx_inc(self):
        self._email_tx += 1

    @wrapt.synchronized(lock)
    @property
    def email_rx(self):
        return self._email_rx

    @wrapt.synchronized(lock)
    def email_rx_inc(self):
        self._email_rx += 1

    @wrapt.synchronized(lock)
    @property
    def email_thread_time(self):
        return self._email_thread_last_time

    @wrapt.synchronized(lock)
    def email_thread_update(self):
        self._email_thread_last_time = datetime.datetime.now()

    @wrapt.synchronized(lock)
    def stats(self):
        now = datetime.datetime.now()
        if self._email_thread_last_time:
            last_update = str(now - self._email_thread_last_time)
        else:
            last_update = "never"

        if self._aprsis_keepalive:
            last_aprsis_keepalive = str(now - self._aprsis_keepalive)
        else:
            last_aprsis_keepalive = "never"

        pm = plugin.PluginManager()
        plugins = pm.get_plugins()
        plugin_stats = {}
        if plugins:
            def full_name_with_qualname(obj):
                return "{}.{}".format(
                    obj.__class__.__module__,
                    obj.__class__.__qualname__,
                )

            for p in plugins:
                plugin_stats[full_name_with_qualname(p)] = {
                    "enabled": p.enabled,
                    "rx": p.rx_count,
                    "tx": p.tx_count,
                    "version": p.version,
                }

        wl = packets.WatchList()
        sl = packets.SeenList()
        pl = packets.PacketList()

        stats = {
            "aprsd": {
                "version": aprsd.__version__,
                "uptime": utils.strfdelta(self.uptime),
                "callsign": CONF.callsign,
                "memory_current": int(self.memory),
                "memory_current_str": utils.human_size(self.memory),
                "memory_peak": int(self.memory_peak),
                "memory_peak_str": utils.human_size(self.memory_peak),
                "threads": self._thread_info,
                "watch_list": wl.get_all(),
                "seen_list": sl.get_all(),
            },
            "aprs-is": {
                "server": str(self.aprsis_server),
                "callsign": CONF.aprs_network.login,
                "last_update": last_aprsis_keepalive,
            },
            "packets": {
                "total_tracked": int(pl.total_tx() + pl.total_rx()),
                "total_sent": int(pl.total_tx()),
                "total_received": int(pl.total_rx()),
                "by_type": self._pkt_cnt,
            },
            "messages": {
                "sent": self._pkt_cnt["MessagePacket"]["tx"],
                "received": self._pkt_cnt["MessagePacket"]["tx"],
                "ack_sent": self._pkt_cnt["AckPacket"]["tx"],
            },
            "email": {
                "enabled": CONF.email_plugin.enabled,
                "sent": int(self._email_tx),
                "received": int(self._email_rx),
                "thread_last_update": last_update,
            },
            "plugins": plugin_stats,
        }
        return stats

    def __str__(self):
        pl = packets.PacketList()
        return (
            "Uptime:{} Msgs TX:{} RX:{} "
            "ACK: TX:{} RX:{} "
            "Email TX:{} RX:{} LastLoop:{} ".format(
                self.uptime,
                pl.total_tx(),
                pl.total_rx(),
                self._pkt_cnt["AckPacket"]["tx"],
                self._pkt_cnt["AckPacket"]["rx"],
                self._email_tx,
                self._email_rx,
                self._email_thread_last_time,
            )
        )



from collections import OrderedDict
from collections.abc import MutableMapping
import logging
import threading

from oslo_config import cfg
import wrapt

from aprsd import stats
from aprsd.packets import seen_list


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class PacketList(MutableMapping):
    _instance = None
    lock = threading.Lock()
    _total_rx: int = 0
    _total_tx: int = 0
    types = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._maxlen = 100
            cls.d = OrderedDict()
        return cls._instance

    @wrapt.synchronized(lock)
    def rx(self, packet):
        """Add a packet that was received."""
        self._total_rx += 1
        self._add(packet)
        ptype = packet.__class__.__name__
        if not ptype in self.types:
            self.types[ptype] = {"tx": 0, "rx": 0}
        self.types[ptype]["rx"] += 1
        seen_list.SeenList().update_seen(packet)
        stats.APRSDStats().rx(packet)

    @wrapt.synchronized(lock)
    def tx(self, packet):
        """Add a packet that was received."""
        self._total_tx += 1
        self._add(packet)
        ptype = packet.__class__.__name__
        if not ptype in self.types:
            self.types[ptype] = {"tx": 0, "rx": 0}
        self.types[ptype]["tx"] += 1
        seen_list.SeenList().update_seen(packet)
        stats.APRSDStats().tx(packet)

    @wrapt.synchronized(lock)
    def add(self, packet):
        self._add(packet)

    def _add(self, packet):
        self[packet.key] = packet

    def copy(self):
        return self.d.copy()

    @property
    def maxlen(self):
        return self._maxlen

    @wrapt.synchronized(lock)
    def find(self, packet):
        return self.get(packet.key)

    def __getitem__(self, key):
        # self.d.move_to_end(key)
        return self.d[key]

    def __setitem__(self, key, value):
        if key in self.d:
            self.d.move_to_end(key)
        elif len(self.d) == self.maxlen:
            self.d.popitem(last=False)
        self.d[key] = value

    def __delitem__(self, key):
        del self.d[key]

    def __iter__(self):
        return self.d.__iter__()

    def __len__(self):
        return len(self.d)

    @wrapt.synchronized(lock)
    def total_rx(self):
        return self._total_rx

    @wrapt.synchronized(lock)
    def total_tx(self):
        return self._total_tx



import queue

# Make these available to anyone importing
# aprsd.threads
from .aprsd import APRSDThread, APRSDThreadList  # noqa: F401
from .keep_alive import KeepAliveThread  # noqa: F401
from .rx import APRSDRXThread, APRSDDupeRXThread, APRSDProcessPacketThread  # noqa: F401


packet_queue = queue.Queue(maxsize=20)



import json
import logging

from oslo_config import cfg
import rpyc
from rpyc.utils.authenticators import AuthenticationError
from rpyc.utils.server import ThreadPoolServer

from aprsd import conf  # noqa: F401
from aprsd import packets, stats, threads
from aprsd.threads import log_monitor


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


def magic_word_authenticator(sock):
    client_ip = sock.getpeername()[0]
    magic = sock.recv(len(CONF.rpc_settings.magic_word)).decode()
    if magic != CONF.rpc_settings.magic_word:
        LOG.error(
            f"wrong magic word passed from {client_ip} "
            "'{magic}' != '{CONF.rpc_settings.magic_word}'",
        )
        raise AuthenticationError(
            f"wrong magic word passed in '{magic}'"
            f" != '{CONF.rpc_settings.magic_word}'",
        )
    return sock, None


class APRSDRPCThread(threads.APRSDThread):
    def __init__(self):
        super().__init__(name="RPCThread")
        self.thread = ThreadPoolServer(
            APRSDService,
            port=CONF.rpc_settings.port,
            protocol_config={"allow_public_attrs": True},
            authenticator=magic_word_authenticator,
        )

    def stop(self):
        if self.thread:
            self.thread.close()
        self.thread_stop = True

    def loop(self):
        # there is no loop as run is blocked
        if self.thread and not self.thread_stop:
            # This is a blocking call
            self.thread.start()


@rpyc.service
class APRSDService(rpyc.Service):
    def on_connect(self, conn):
        # code that runs when a connection is created
        # (to init the service, if needed)
        LOG.info("RPC Client Connected")
        self._conn = conn

    def on_disconnect(self, conn):
        # code that runs after the connection has already closed
        # (to finalize the service, if needed)
        LOG.info("RPC Client Disconnected")
        self._conn = None

    @rpyc.exposed
    def get_stats(self):
        stat = stats.APRSDStats()
        stats_dict = stat.stats()
        return_str = json.dumps(stats_dict, indent=4, sort_keys=True, default=str)
        return return_str

    @rpyc.exposed
    def get_stats_obj(self):
        return stats.APRSDStats()

    @rpyc.exposed
    def get_packet_list(self):
        return packets.PacketList()

    @rpyc.exposed
    def get_packet_track(self):
        return packets.PacketTrack()

    @rpyc.exposed
    def get_watch_list(self):
        return packets.WatchList()

    @rpyc.exposed
    def get_seen_list(self):
        return packets.SeenList()

    @rpyc.exposed
    def get_log_entries(self):
        entries = log_monitor.LogEntries().get_all_and_purge()
        return json.dumps(entries, default=str)



import unittest
from unittest import mock

from aprsd import packets
from aprsd.packets import core

from . import fake


class TestPluginBase(unittest.TestCase):

    def _fake_dict(
        self,
        from_call=fake.FAKE_FROM_CALLSIGN,
        to_call=fake.FAKE_TO_CALLSIGN,
        message=None,
        msg_number=None,
        message_format=core.PACKET_TYPE_MESSAGE,
    ):
        packet_dict = {
            "from": from_call,
            "addresse": to_call,
            "to": to_call,
            "format": message_format,
            "raw": "",
        }

        if message:
            packet_dict["message_text"] = message

        if msg_number:
            packet_dict["msgNo"] = str(msg_number)

        return packet_dict

    def test_packet_construct(self):
        pkt = packets.Packet(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
        )

        self.assertEqual(fake.FAKE_FROM_CALLSIGN, pkt.from_call)
        self.assertEqual(fake.FAKE_TO_CALLSIGN, pkt.to_call)

    def test_packet_get_attr(self):
        pkt = packets.Packet(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
        )

        self.assertEqual(
            fake.FAKE_FROM_CALLSIGN,
            pkt.get("from_call"),
        )

    def test_packet_factory(self):
        pkt_dict = self._fake_dict()
        pkt = packets.Packet.factory(pkt_dict)

        self.assertIsInstance(pkt, packets.MessagePacket)
        self.assertEqual(fake.FAKE_FROM_CALLSIGN, pkt.from_call)
        self.assertEqual(fake.FAKE_TO_CALLSIGN, pkt.to_call)
        self.assertEqual(fake.FAKE_TO_CALLSIGN, pkt.addresse)

        pkt_dict["symbol"] = "_"
        pkt_dict["weather"] = {
            "wind_gust": 1.11,
            "temperature": 32.01,
            "humidity": 85,
            "pressure": 1095.12,
            "comment": "Home!",
        }
        pkt_dict["format"] = core.PACKET_TYPE_UNCOMPRESSED
        pkt = packets.Packet.factory(pkt_dict)
        self.assertIsInstance(pkt, packets.WeatherPacket)

    @mock.patch("aprsd.packets.core.GPSPacket._build_time_zulu")
    def test_packet_format_rain_1h(self, mock_time_zulu):

        mock_time_zulu.return_value = "221450"

        wx = packets.WeatherPacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
            timestamp=1671721164.1112509,
        )
        wx.prepare()

        expected = "KFAKE>KMINE,WIDE1-1,WIDE2-1:@221450z0.0/0.0_000/000g000t000r000p000P000h00b00000"
        self.assertEqual(expected, wx.raw)
        rain_location = 59
        self.assertEqual(rain_location, wx.raw.find("r000"))

        wx.rain_1h = 1.11
        wx.prepare()
        expected = "KFAKE>KMINE,WIDE1-1,WIDE2-1:@221450z0.0/0.0_000/000g000t000r111p000P000h00b00000"
        self.assertEqual(expected, wx.raw)

        wx.rain_1h = 0.01
        wx.prepare()
        expected = "KFAKE>KMINE,WIDE1-1,WIDE2-1:@221450z0.0/0.0_000/000g000t000r001p000P000h00b00000"
        self.assertEqual(expected, wx.raw)



import logging
import threading

import wrapt

from aprsd import threads
from aprsd.log import log


LOG = logging.getLogger("APRSD")


class LogEntries:
    entries = []
    lock = threading.Lock()
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @wrapt.synchronized(lock)
    def add(self, entry):
        self.entries.append(entry)

    @wrapt.synchronized(lock)
    def get_all_and_purge(self):
        entries = self.entries.copy()
        self.entries = []
        return entries

    @wrapt.synchronized(lock)
    def __len__(self):
        return len(self.entries)


class LogMonitorThread(threads.APRSDThread):

    def __init__(self):
        super().__init__("LogMonitorThread")

    def loop(self):
        try:
            record = log.logging_queue.get(block=True, timeout=2)
            if isinstance(record, list):
                for item in record:
                    entry = self.json_record(item)
                    LogEntries().add(entry)
            else:
                entry = self.json_record(record)
                LogEntries().add(entry)
        except Exception:
            # Just ignore thi
            pass

        return True

    def json_record(self, record):
        entry = {}
        entry["filename"] = record.filename
        entry["funcName"] = record.funcName
        entry["levelname"] = record.levelname
        entry["lineno"] = record.lineno
        entry["module"] = record.module
        entry["name"] = record.name
        entry["pathname"] = record.pathname
        entry["process"] = record.process
        entry["processName"] = record.processName
        if hasattr(record, "stack_info"):
            entry["stack_info"] = record.stack_info
        else:
            entry["stack_info"] = None
        entry["thread"] = record.thread
        entry["threadName"] = record.threadName
        entry["message"] = record.getMessage()
        return entry



import json
import logging

from oslo_config import cfg
import rpyc

from aprsd import conf  # noqa
from aprsd import rpc


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class RPCClient:
    _instance = None
    _rpc_client = None

    ip = None
    port = None
    magic_word = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, ip=None, port=None, magic_word=None):
        if ip:
            self.ip = ip
        else:
            self.ip = CONF.rpc_settings.ip
        if port:
            self.port = int(port)
        else:
            self.port = CONF.rpc_settings.port
        if magic_word:
            self.magic_word = magic_word
        else:
            self.magic_word = CONF.rpc_settings.magic_word
        self._check_settings()
        self.get_rpc_client()

    def _check_settings(self):
        if not CONF.rpc_settings.enabled:
            LOG.warning("RPC is not enabled, no way to get stats!!")

        if self.magic_word == conf.common.APRSD_DEFAULT_MAGIC_WORD:
            LOG.warning("You are using the default RPC magic word!!!")
            LOG.warning("edit aprsd.conf and change rpc_settings.magic_word")

        LOG.debug(f"RPC Client: {self.ip}:{self.port} {self.magic_word}")

    def _rpyc_connect(
            self, host, port, service=rpyc.VoidService,
            config={}, ipv6=False,
            keepalive=False, authorizer=None, ):

        LOG.info(f"Connecting to RPC host '{host}:{port}'")
        try:
            s = rpc.AuthSocketStream.connect(
                host, port, ipv6=ipv6, keepalive=keepalive,
                authorizer=authorizer,
            )
            return rpyc.utils.factory.connect_stream(s, service, config=config)
        except ConnectionRefusedError:
            LOG.error(f"Failed to connect to RPC host '{host}:{port}'")
            return None

    def get_rpc_client(self):
        if not self._rpc_client:
            self._rpc_client = self._rpyc_connect(
                self.ip,
                self.port,
                authorizer=lambda sock: sock.send(self.magic_word.encode()),
            )
        return self._rpc_client

    def get_stats_dict(self):
        cl = self.get_rpc_client()
        result = {}
        if not cl:
            return result

        try:
            rpc_stats_dict = cl.root.get_stats()
            result = json.loads(rpc_stats_dict)
        except EOFError:
            LOG.error("Lost connection to RPC Host")
            self._rpc_client = None
        return result

    def get_stats(self):
        cl = self.get_rpc_client()
        result = {}
        if not cl:
            return result

        try:
            result = cl.root.get_stats_obj()
        except EOFError:
            LOG.error("Lost connection to RPC Host")
            self._rpc_client = None
        return result

    def get_packet_track(self):
        cl = self.get_rpc_client()
        result = None
        if not cl:
            return result
        try:
            result = cl.root.get_packet_track()
        except EOFError:
            LOG.error("Lost connection to RPC Host")
            self._rpc_client = None
        return result

    def get_packet_list(self):
        cl = self.get_rpc_client()
        result = None
        if not cl:
            return result
        try:
            result = cl.root.get_packet_list()
        except EOFError:
            LOG.error("Lost connection to RPC Host")
            self._rpc_client = None
        return result

    def get_watch_list(self):
        cl = self.get_rpc_client()
        result = None
        if not cl:
            return result
        try:
            result = cl.root.get_watch_list()
        except EOFError:
            LOG.error("Lost connection to RPC Host")
            self._rpc_client = None
        return result

    def get_seen_list(self):
        cl = self.get_rpc_client()
        result = None
        if not cl:
            return result
        try:
            result = cl.root.get_seen_list()
        except EOFError:
            LOG.error("Lost connection to RPC Host")
            self._rpc_client = None
        return result

    def get_log_entries(self):
        cl = self.get_rpc_client()
        result = None
        if not cl:
            return result
        try:
            result_str = cl.root.get_log_entries()
            result = json.loads(result_str)
        except EOFError:
            LOG.error("Lost connection to RPC Host")
            self._rpc_client = None
        return result



from unittest import mock

from oslo_config import cfg

from aprsd import packets
from aprsd.packets import tracker
from aprsd.plugins import query as query_plugin

from .. import fake, test_plugin


CONF = cfg.CONF


class TestQueryPlugin(test_plugin.TestPlugin):
    @mock.patch("aprsd.packets.tracker.PacketTrack.flush")
    def test_query_flush(self, mock_flush):
        packet = fake.fake_packet(message="!delete")
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        CONF.save_enabled = True
        CONF.query_plugin.callsign = fake.FAKE_FROM_CALLSIGN
        query = query_plugin.QueryPlugin()
        query.enabled = True

        expected = "Deleted ALL pending msgs."
        actual = query.filter(packet)
        mock_flush.assert_called_once()
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.packets.tracker.PacketTrack.restart_delayed")
    def test_query_restart_delayed(self, mock_restart):
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        CONF.save_enabled = True
        CONF.query_plugin.callsign = fake.FAKE_FROM_CALLSIGN
        track = tracker.PacketTrack()
        track.data = {}
        packet = fake.fake_packet(message="!4")
        query = query_plugin.QueryPlugin()

        expected = "No pending msgs to resend"
        actual = query.filter(packet)
        mock_restart.assert_not_called()
        self.assertEqual(expected, actual)
        mock_restart.reset_mock()

        # add a message
        pkt = packets.MessagePacket(
            from_call=self.fromcall,
            to_call="testing",
            msgNo=self.ack,
        )
        track.add(pkt)
        actual = query.filter(packet)
        mock_restart.assert_called_once()



#
#  Used to fetch the stats url and determine if
#  aprsd server is 'healthy'
#
#
# python included libs
import datetime
import logging
import sys

import click
from oslo_config import cfg
from rich.console import Console

import aprsd
from aprsd import cli_helper, utils
from aprsd import conf  # noqa
# local imports here
from aprsd.main import cli
from aprsd.rpc import client as aprsd_rpc_client


# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
CONF = cfg.CONF
LOG = logging.getLogger("APRSD")
console = Console()


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "--timeout",
    show_default=True,
    default=3,
    help="How long to wait for healtcheck url to come back",
)
@click.pass_context
@cli_helper.process_standard_options
def healthcheck(ctx, timeout):
    """Check the health of the running aprsd server."""
    console.log(f"APRSD HealthCheck version: {aprsd.__version__}")
    if not CONF.rpc_settings.enabled:
        LOG.error("Must enable rpc_settings.enabled to use healthcheck")
        sys.exit(-1)
    if not CONF.rpc_settings.ip:
        LOG.error("Must enable rpc_settings.ip to use healthcheck")
        sys.exit(-1)
    if not CONF.rpc_settings.magic_word:
        LOG.error("Must enable rpc_settings.magic_word to use healthcheck")
        sys.exit(-1)

    with console.status(f"APRSD HealthCheck version: {aprsd.__version__}") as status:
        try:
            status.update(f"Contacting APRSD via RPC {CONF.rpc_settings.ip}")
            stats = aprsd_rpc_client.RPCClient().get_stats_dict()
        except Exception as ex:
            console.log(f"Failed to fetch healthcheck : '{ex}'")
            sys.exit(-1)
        else:
            if not stats:
                console.log("No stats from aprsd")
                sys.exit(-1)
            email_thread_last_update = stats["email"]["thread_last_update"]

            if email_thread_last_update != "never":
                delta = utils.parse_delta_str(email_thread_last_update)
                d = datetime.timedelta(**delta)
                max_timeout = {"hours": 0.0, "minutes": 5, "seconds": 0}
                max_delta = datetime.timedelta(**max_timeout)
                if d > max_delta:
                    console.log(f"Email thread is very old! {d}")
                    sys.exit(-1)

            aprsis_last_update = stats["aprs-is"]["last_update"]
            delta = utils.parse_delta_str(aprsis_last_update)
            d = datetime.timedelta(**delta)
            max_timeout = {"hours": 0.0, "minutes": 5, "seconds": 0}
            max_delta = datetime.timedelta(**max_timeout)
            if d > max_delta:
                LOG.error(f"APRS-IS last update is very old! {d}")
                sys.exit(-1)

            sys.exit(0)



import logging
import select
import threading

import aprslib
from aprslib import is_py3
from aprslib.exceptions import (
    ConnectionDrop, ConnectionError, GenericError, LoginError, ParseError,
    UnknownFormat,
)
import wrapt

import aprsd
from aprsd import stats
from aprsd.packets import core


LOG = logging.getLogger("APRSD")


class Aprsdis(aprslib.IS):
    """Extend the aprslib class so we can exit properly."""

    # flag to tell us to stop
    thread_stop = False

    # timeout in seconds
    select_timeout = 1
    lock = threading.Lock()

    def stop(self):
        self.thread_stop = True
        LOG.info("Shutdown Aprsdis client.")

    @wrapt.synchronized(lock)
    def send(self, packet: core.Packet):
        """Send an APRS Message object."""
        self.sendall(packet.raw)

    def is_alive(self):
        """If the connection is alive or not."""
        return self._connected

    def _socket_readlines(self, blocking=False):
        """
        Generator for complete lines, received from the server
        """
        try:
            self.sock.setblocking(0)
        except OSError as e:
            self.logger.error(f"socket error when setblocking(0): {str(e)}")
            raise aprslib.ConnectionDrop("connection dropped")

        while not self.thread_stop:
            short_buf = b""
            newline = b"\r\n"

            # set a select timeout, so we get a chance to exit
            # when user hits CTRL-C
            readable, writable, exceptional = select.select(
                [self.sock],
                [],
                [],
                self.select_timeout,
            )
            if not readable:
                if not blocking:
                    break
                else:
                    continue

            try:
                short_buf = self.sock.recv(4096)

                # sock.recv returns empty if the connection drops
                if not short_buf:
                    if not blocking:
                        # We could just not be blocking, so empty is expected
                        continue
                    else:
                        self.logger.error("socket.recv(): returned empty")
                        raise aprslib.ConnectionDrop("connection dropped")
            except OSError as e:
                # self.logger.error("socket error on recv(): %s" % str(e))
                if "Resource temporarily unavailable" in str(e):
                    if not blocking:
                        if len(self.buf) == 0:
                            break

            self.buf += short_buf

            while newline in self.buf:
                line, self.buf = self.buf.split(newline, 1)

                yield line

    def _send_login(self):
        """
        Sends login string to server
        """
        login_str = "user {0} pass {1} vers github.com/craigerl/aprsd {3}{2}\r\n"
        login_str = login_str.format(
            self.callsign,
            self.passwd,
            (" filter " + self.filter) if self.filter != "" else "",
            aprsd.__version__,
        )

        self.logger.debug("Sending login information")

        try:
            self._sendall(login_str)
            self.sock.settimeout(5)
            test = self.sock.recv(len(login_str) + 100)
            if is_py3:
                test = test.decode("latin-1")
            test = test.rstrip()

            self.logger.debug("Server: '%s'", test)

            if not test:
                raise LoginError(f"Server Response Empty: '{test}'")

            _, _, callsign, status, e = test.split(" ", 4)
            s = e.split(",")
            if len(s):
                server_string = s[0].replace("server ", "")
            else:
                server_string = e.replace("server ", "")

            if callsign == "":
                raise LoginError("Server responded with empty callsign???")
            if callsign != self.callsign:
                raise LoginError(f"Server: {test}")
            if status != "verified," and self.passwd != "-1":
                raise LoginError("Password is incorrect")

            if self.passwd == "-1":
                self.logger.info("Login successful (receive only)")
            else:
                self.logger.info("Login successful")

            self.logger.info(f"Connected to {server_string}")
            self.server_string = server_string
            stats.APRSDStats().set_aprsis_server(server_string)

        except LoginError as e:
            self.logger.error(str(e))
            self.close()
            raise
        except Exception as e:
            self.close()
            self.logger.error(f"Failed to login '{e}'")
            self.logger.exception(e)
            raise LoginError("Failed to login")

    def consumer(self, callback, blocking=True, immortal=False, raw=False):
        """
        When a position sentence is received, it will be passed to the callback function

        blocking: if true (default), runs forever, otherwise will return after one sentence
                  You can still exit the loop, by raising StopIteration in the callback function

        immortal: When true, consumer will try to reconnect and stop propagation of Parse exceptions
                  if false (default), consumer will return

        raw: when true, raw packet is passed to callback, otherwise the result from aprs.parse()
        """

        if not self._connected:
            raise ConnectionError("not connected to a server")

        line = b""

        while True and not self.thread_stop:
            try:
                for line in self._socket_readlines(blocking):
                    if line[0:1] != b"#":
                        if raw:
                            callback(line)
                        else:
                            callback(self._parse(line))
                    else:
                        self.logger.debug("Server: %s", line.decode("utf8"))
                        stats.APRSDStats().set_aprsis_keepalive()
            except ParseError as exp:
                self.logger.log(
                    11,
                    "%s\n    Packet: %s",
                    exp,
                    exp.packet,
                )
            except UnknownFormat as exp:
                self.logger.log(
                    9,
                    "%s\n    Packet: %s",
                    exp,
                    exp.packet,
                )
            except LoginError as exp:
                self.logger.error("%s: %s", exp.__class__.__name__, exp)
            except (KeyboardInterrupt, SystemExit):
                raise
            except (ConnectionDrop, ConnectionError):
                self.close()

                if not immortal:
                    raise
                else:
                    self.connect(blocking=blocking)
                    continue
            except GenericError:
                pass
            except StopIteration:
                break
            except Exception:
                self.logger.error("APRS Packet: %s", line)
                raise

            if not blocking:
                break
